from __future__ import annotations

from random import randint
from typing import Optional, List, Union, Tuple, Literal
from urllib.parse import urlencode

import numpy as np
import pandas as pd

import scipy.stats

import okama.portfolios.dcf as dcf
from okama import settings
from okama.common.helpers.rebalancing import Rebalance
from okama.common import make_asset_list, validators
from okama.common.helpers import helpers, ratios


class Portfolio(make_asset_list.ListMaker):
    """
    Investment portfolio.

    An investment portfolio is a type of financial asset (same as stocks, ETFs, mutual funds, currencies, etc.).
    The arguments are similar to `AssetList`, but `Portfolio` additionally has:

    - `weights`
    - `rebalancing_strategy`
    - `symbol`

    Portfolio is defined by the investment strategy, which includes:

    - asset allocation (financial assets and their proportions in the portfolio)
    - the rebalancing strategy (`rebalancing_strategy` parameter)

    Rebalancing is the action of bringing the portfolio that has deviated away from the original target asset
    allocation back into line. After rebalancing, the portfolio assets have the original weights.

    Parameters
    ----------
    assets : list, default None
        List of assets. Could include tickers or asset-like objects (`Asset`, `Portfolio`).
        If None, a single-asset list with a default ticker is used.
    first_date : str, default None
        First date of monthly return time series.
        If None, the first date is calculated automatically as the oldest available date for the listed assets.
    last_date : str, default None
        Last date of monthly return time series.
        If None, the last date is calculated automatically as the newest available date for the listed assets.
    ccy : str, default 'USD'
        Base currency for the list of assets. All risk metrics and returns are adjusted to the base currency.
    inflation : bool, default True
        Defines whether to take inflation data into account in the calculations.
        Including inflation could limit available data (`first_date`, `last_date`)
        as the inflation data is usually published with a one-month delay.
        With `inflation=False`, some properties like real return are not available.
    weights : list of float, default None
        List of asset weights.
        The weight of an asset is the percent of an investment portfolio that corresponds to the asset.
        If None, an equally weighted portfolio is created (all weights are equal).
    rebalancing_strategy : Rebalance, default Rebalance(period="month")
        Rebalancing strategy for an investment portfolio. The strategy is defined by:

        - period (rebalancing frequency): predetermined time intervals when the investor rebalances the portfolio.
          If "none", asset weights are not rebalanced.
        - abs_deviation: the absolute deviation allowed for the asset weights in the portfolio.
        - rel_deviation: the relative deviation allowed for the asset weights in the portfolio.
    symbol : str, default None
        Text symbol of portfolio. It is similar to tickers but have a namespace information.
        Portfolio symbol must end with .PF (all_weather_portfolio.PF).
        If None, a random symbol is generated (portfolio_7802.PF).
    """

    def __init__(
        self,
        assets: Optional[List[str]] = None,
        *,
        first_date: Optional[str] = None,
        last_date: Optional[str] = None,
        ccy: str = "USD",
        inflation: bool = True,
        weights: Optional[List[float]] = None,
        rebalancing_strategy: Rebalance = Rebalance(period="month"),
        symbol: str = None,
    ):
        super().__init__(
            assets,
            first_date=first_date,
            last_date=last_date,
            ccy=ccy,
            inflation=inflation,
        )
        self.weights = weights
        self.assets_weights = dict(zip(self.symbols, self.weights))
        self.rebalancing_strategy = rebalancing_strategy
        self.symbol = symbol or f"portfolio_{randint(1000, 9999)}.PF"
        self._ror = pd.DataFrame(dtype=float)
        self.dcf = dcf.PortfolioDCF(self)

    def __repr__(self):
        dic = {
            "symbol": self.symbol,
            "assets": self.symbols,
            "weights": self.weights,
            "rebalancing_period": self.rebalancing_strategy.period,
            "rebalancing_abs_deviation": self.rebalancing_strategy.abs_deviation,
            "rebalancing_rel_deviation": self.rebalancing_strategy.rel_deviation,
            "currency": self.currency,
            "inflation": self.inflation if hasattr(self, "inflation") else "None",
            "first_date": self.first_date.strftime("%Y-%m"),
            "last_date": self.last_date.strftime("%Y-%m"),
            "period_length": self._pl_txt,
        }
        return repr(pd.Series(dic))

    def _add_inflation(self):
        if hasattr(self, "inflation"):
            return pd.concat([self.ror, self.inflation_ts], axis=1, join="inner", copy="false")
        else:
            return self.ror

    def _clear_cache(self):
        self._ror = pd.DataFrame(dtype=float)
        try:
            self.dcf._wealth_index_fv = pd.DataFrame()
            self.dcf._monte_carlo_wealth_fv = pd.DataFrame()
            self.dcf._cash_flow_fv = pd.DataFrame()
        except AttributeError:
            pass

    # todo: add setters for dates and ccy

    @property
    def weights(self) -> Union[list, tuple]:
        """
        Assets weights in portfolio.

        If not defined equal weights are used for each asset.

        Weights must be a list (or tuple) of float values.

        Returns
        -------
        list or tuple
            Values for the weights of assets in portfolio.

        Examples
        --------
        >>> x = ok.Portfolio(['SPY.US', 'BND.US'])
        >>> x.weights
        [0.5, 0.5]
        """
        return self._weights

    @weights.setter
    def weights(self, weights: Optional[List[float]]):
        if weights is None:
            # Equally weighted portfolio
            n = len(self.symbols)  # number of assets
            weights = list(np.repeat(1 / n, n))
        else:
            [validators.validate_real("weight", weight) for weight in weights]
            helpers.Frame.weights_sum_is_one(weights)
            if len(weights) != len(set(self.symbols)):
                raise ValueError(
                    f"Number of tickers ({len(set(self.symbols))}) should be equal "
                    f"to the weights number ({len(weights)})"
                )
        self._clear_cache()
        self._weights = weights

    @property
    def weights_ts(self) -> pd.DataFrame:
        """
        Calculate assets weights time series considering rebalancing strategy.

        The weights of assets in Portfolio are not constant if rebalancing_period is different from 'month'.

        Returns
        -------
        DataFrame
            Weights of assets time series.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> reb_period = ok.Rebalance(period="none")  # The Portfolio is not rebalanced.
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US'], weights=[0.5, 0.5], rebalancing_strategy=reb_period)
        >>> pf.weights_ts.plot()
        >>> plt.show()

        The weights of assets time series will differ significantly if the portfolio rebalancing_period is 1 year.

        >>> pf.rebalancing_strategy = ok.Rebalance(period="year")  # set a new rebalancing period
        >>> pf.weights_ts.plot()
        >>> plt.show()
        """
        if self._condition_for_rebalancing:
            return self.rebalancing_strategy.assets_weights_ts(
                ror=self.assets_ror,
                target_weights=self.weights,
            )
        # Fast calculation
        values = np.tile(self.weights, (self.ror.shape[0], 1))
        return pd.DataFrame(values, index=self.ror.index, columns=self.symbols)

    @property
    def rebalancing_events(self) -> pd.DataFrame:
        """
        Time series with the dates of rebalancing events.

        Each event has the type of rebalancing event:
        - calendar (calendar event)
        - abs (rebalancing by absolute deviation)
        - rel (rebalancing by relative deviation)

        Returns
        -------
        DataFrame
            Dates of rebalancing events time series.

        Examples
        --------
        >>> pf = ok.Portfolio(
        ...     assets=["SPY.US", "AGG.US"],
        ...     weights=[0.5, 0.5],
        ...     rebalancing_strategy=ok.Rebalance(period="year", abs_deviation=0.05)
        ... )
        >>> pf.rebalancing_events
        date
        2004-05    calendar
        2005-05    calendar
        2006-05    calendar
        2007-05    calendar
        2008-05    calendar
                     ...
        2020-05    calendar
        2020-11         abs
        2021-05    calendar
        2022-05    calendar
        2023-05    calendar
        Name: event, Length: 21, dtype: object
        """
        return self.rebalancing_strategy.wealth_ts(target_weights=self.weights, ror=self.assets_ror).events

    @property
    def rebalancing_strategy(self) -> Rebalance:
        """
        Return rebalancing strategy of the portfolio.

        Rebalancing is the process by which an investor restores their portfolio to its target allocation
        by selling and buying assets. After rebalancing all the assets have original weights.

        Rebalancing period (rebalancing frequency) is predetermined time intervals when
        the investor rebalances the portfolio.

        Returns
        -------
        Rebalance
            Portfolio rebalancing strategy.

        Examples
        --------
        >>> pf = ok.Portfolio(
        ...     assets=["SPY.US", "AGG.US"],
        ...     weights=[0.5, 0.5],
        ...     rebalancing_strategy=ok.Rebalance(period="year", abs_deviation=0.05)
        ... )
        >>> pf.rebalancing_strategy
        Rebalance(period='year', abs_deviation=0.05, rel_deviation=None)
        """
        return self._rebalancing_strategy

    @rebalancing_strategy.setter
    def rebalancing_strategy(self, rebalancing_strategy: Rebalance):
        if isinstance(rebalancing_strategy, Rebalance):
            self._clear_cache()
            self._rebalancing_strategy = rebalancing_strategy
        else:
            raise ValueError("rebalancing_strategy must be of type Rebalance")

    @property
    def _condition_for_rebalancing(self) -> bool:
        """
        Verify whether assets weights are constant
        The weights are constant only if the period is 'month' and no conditional rebalancing.
        """
        return (
            self.rebalancing_strategy.period != "month"
            or self.rebalancing_strategy.abs_deviation is not None
            or self.rebalancing_strategy.rel_deviation is not None
        )

    @property
    def symbol(self) -> str:
        """
        Return a text symbol of portfolio.

        Symbols are similar to tickers but have a namespace information:

        * SPY.US is a symbol
        * SPY is a ticker

        Portfolios have '.PF' as a namespace.

        Returns
        -------
        str
            Text symbol of the portfolio.

        Examples
        --------
        >>> p = ok.Portfolio()
        >>> p.symbol  # a randomly generated symbol will be shown
        'portfolio_5312.PF'

        >>> p.symbol = 'spy_portfolo.PF'  # The symbol can be customized after initialization

        New Portfolio can have a custom symbol.

        >>> p = ok.Portfolio(symbol='aggressive.PF')
        >>> p.symbol
        'aggressive.PF'
        """
        return self._symbol

    @symbol.setter
    def symbol(self, text_symbol: str):
        if isinstance(text_symbol, str) and text_symbol.endswith(".PF"):
            if " " in text_symbol:
                raise ValueError("portfolio text symbol should not have whitespace characters.")
            self._clear_cache()
            self._symbol = text_symbol
        else:
            raise ValueError('portfolio symbol must be a string ending with ".PF" namespace.')

    @property
    def name(self) -> str:
        """
        Return text name of portfolio.

        For Portfolio name is equal to symbol.

        Returns
        -------
        str
            Text name of the portfolio.

        >>> p = ok.Portfolio()
        >>> p.name
        'portfolio_5312.PF'
        """
        return self.symbol

    @property
    def ror(self) -> pd.Series:
        """
        Calculate monthly rate of return time series for portfolio considering rebalancing strategy.

        Returns
        -------
        Series
            Rate of return monthly time series for the portfolio.

        Examples
        --------
        >>> pf = ok.Portfolio(first_date='2020-01', last_date='2020-12')
        >>> pf.ror
        Date
        2020-01   -0.0004
        2020-02   -0.0792
        2020-03   -0.1249
        2020-04    0.1270
        2020-05    0.0476
        2020-06    0.0177
        2020-07    0.0589
        2020-08    0.0698
        2020-09   -0.0374
        2020-10   -0.0249
        2020-11    0.1088
        2020-12    0.0370
        Freq: M, Name: portfolio_4669.PF, dtype: float64

        >>> import matplotlib.pyplot as plt
        >>> pf.ror.plot(kind='bar')
        >>> plt.show()
        """
        if self._ror.empty:
            if not self._condition_for_rebalancing:
                # Fast calculation
                s = helpers.Frame.get_portfolio_return_ts(self.weights, self.assets_ror)
            else:
                s = self.rebalancing_strategy.return_ror_ts(self.weights, self.assets_ror)
            s.rename(self.symbol, inplace=True)
            self._ror = s
        return self._ror

    @property
    def wealth_index(self) -> pd.DataFrame:
        """
        Calculate wealth index time series for the portfolio and accumulated inflation.

        Wealth index (Cumulative Wealth Index) is a time series that presents the value of portfolio over
        historical time period. Accumulated inflation time series is added if `inflation=True` in the Portfolio.

        Wealth index is obtained from the accumulated return multiplied by the initial investments.
        That is: 1000 * (Acc_Return + 1)
        Initial investments are taken as 1000 units of the Portfolio base currency.

        Returns
        -------
        DataFrame
            Time series of wealth index values for portfolio and accumulated inflation.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> x = ok.Portfolio(['SPY.US', 'BND.US'])
        >>> x.wealth_index.plot()
        >>> plt.show()
        """
        df = self._add_inflation()
        iv = settings.DEFAULT_INITIAL_INVESTMENT
        df = helpers.Frame.get_wealth_indexes(ror=df, initial_amount=iv)
        df = self._make_df_if_series(df)
        return df

    def _make_df_if_series(self, ts):
        if isinstance(ts, pd.Series):  # should always return a DataFrame
            ts = ts.to_frame()
            ts.rename({1: self.symbol}, axis="columns", inplace=True)
        return ts

    @property
    def wealth_index_with_assets(self) -> pd.DataFrame:
        """
        Calculate wealth index time series for the portfolio, all assets and accumulated inflation.

        Cash flows are not taken into account.

        Wealth index (Cumulative Wealth Index) is a time series that presents the value of portfolio over
        historical time period. Accumulated inflation time series is added if `inflation=True` in the Portfolio.

        Wealth index is obtained from the accumulated return multiplied by the initial investments.
        initial_amount_pv * (Acc_Return + 1)

        Returns
        -------
        DataFrame
            Time series of wealth index values for portfolio, each asset and accumulated inflation.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['VOO.US', 'GLD.US'], weights=[0.8, 0.2])
        >>> pf.wealth_index_with_assets.plot()
        >>> plt.show()
        """
        ls = [self.ror, self.assets_ror]
        if hasattr(self, "inflation"):
            ls.append(self.inflation_ts)
        df = pd.concat(ls, axis=1, join="inner", copy="false")
        return helpers.Frame.get_wealth_indexes(df)

    @property
    def mean_return_monthly(self) -> float:
        """
        Calculate monthly mean return (arithmetic mean) for the portfolio rate of return time series.

        Mean return calculated for the full history period.

        Returns
        -------
        float
            Mean return value.

        Examples
        --------
        >>> pf = ok.Portfolio(['ISF.LSE', 'XGLE.LSE'], weights=[0.6, 0.4], ccy='GBP')
        >>> pf.mean_return_monthly
        0.0001803312727272665
        """
        return self.ror.mean()

    @property
    def mean_return_annual(self) -> float:
        """
        Calculate annualized mean return (arithmetic mean) for the portfolio rate of return time series.

        Mean return calculated for the full history period.

        Returns
        -------
        float
            Mean return value.

        Examples
        --------
        >>> pf = ok.Portfolio(['XCS6.XETR', 'PHAU.LSE'], weights=[0.85, 0.15], ccy='USD')
        >>> pf.names
        {'XCS6.XETR': 'Xtrackers MSCI China UCITS ETF 1C', 'PHAU.LSE': 'WisdomTree Physical Gold'}
        >>> pf.mean_return_annual
        0.09005826844072184
        """
        return self.mean_return_monthly * settings._MONTHS_PER_YEAR

    def annual_return_ts(self, return_type: Literal["cagr", "arithmetic_mean"] = "cagr") -> pd.Series:
        """
        Calculate annual rate of return time series for portfolio.

        Rate of return is calculated for each calendar year.

        Parameters
        ----------
        return_type : {'cagr', 'arithmetic_mean'}, default 'cagr'
            Method used to calculate annual returns.

        Returns
        -------
        Series
            Calendar annual rate of return time series.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['VOO.US', 'AGG.US'], weights=[0.4, 0.6])
        >>> pf.annual_return_ts().plot(kind='bar')
        >>> plt.show()

        Plot annual returns for portfolio with EUR as the base currency.

        >>> pf = ok.Portfolio(['VOO.US', 'AGG.US'], weights=[0.4, 0.6], ccy='EUR')
        >>> pf.annual_return_ts().plot(kind='bar')
        >>> plt.show()
        """
        return helpers.Frame.get_annual_return_ts_from_monthly(self.ror, return_type)

    def get_cagr(self, period: Optional[int] = None, real: bool = False) -> pd.Series:
        """
        Calculate portfolio Compound Annual Growth Rate (CAGR) for a given trailing period.

        Compound annual growth rate (CAGR) is the rate of return that would be required for an investment to grow from
        its initial to its final value, assuming all incomes were reinvested.

        Inflation adjusted annualized returns (real CAGR) are shown with `real=True` option.

        Annual inflation value is calculated for the same period if `inflation=True` in the `Portfolio`.

        Parameters
        ----------
        period : int, default None
            CAGR trailing period in years. If None, use the full available period.
        real : bool, default False
            CAGR is adjusted for inflation (real CAGR) if True.
            Portfolio should be initiated with `inflation=True` for real CAGR.

        Returns
        -------
        Series
            Portfolio CAGR value and annualized inflation (optional).

        Notes
        -----
        CAGR is not defined for periods less than 1 year (NaN values are returned).

        Examples
        --------
        >>> pf = ok.Portfolio(['XCS6.XETR', 'PHAU.LSE'], weights=[0.85, 0.15], ccy='USD')
        >>> pf.names
        {'XCS6.XETR': 'Xtrackers MSCI China UCITS ETF 1C', 'PHAU.LSE': 'WisdomTree Physical Gold'}

        To get inflation adjusted return (real annualized return) add `real=True` option:

        >>> pf.get_cagr(period=5, real=True)
        portfolio_5625.PF    0.121265
        dtype: float64
        """
        # TODO: add option assets=False
        ts = self._add_inflation()
        df = self._make_df_if_series(ts)
        dt0 = self.last_date
        if period is None:
            dt = self.first_date
        else:
            self._validate_period(period)
            dt = helpers.Date.subtract_years(dt0, period)
        cagr = helpers.Frame.get_cagr(df[dt:])
        if real:
            if not hasattr(self, "inflation"):
                raise ValueError("Real CAGR is not defined. Set inflation=True in Portfolio to calculate it.")
            mean_inflation = helpers.Frame.get_cagr(self.inflation_ts[dt:])
            cagr = (1.0 + cagr) / (1.0 + mean_inflation) - 1.0
            cagr.drop(self.inflation, inplace=True)
        return cagr

    def get_rolling_cagr(self, window: int = 12, real: bool = False) -> pd.DataFrame:
        """
        Calculate rolling CAGR (Compound Annual Growth Rate) for the portfolio.

        Parameters
        ----------
        window : int, default 12
            Size of the moving window in months. Window size should be at least 12 months for CAGR.
        real : bool, default False
            CAGR is adjusted for inflation (real CAGR) if True.
            Portfolio should be initiated with Inflation=True for real CAGR.

        Returns
        -------
        DataFrame
            Time series of rolling CAGR and mean inflation (optionally).

        Notes
        -----
        CAGR is not defined for periods less than 1 year (NaN values are returned).

        Examples
        --------
        >>> x = ok.Portfolio(['SPY.US', 'BND.US'], ccy='EUR', inflation=True)
        >>> x.get_rolling_cagr(window=5*12, real=True)
                 portfolio_...
        date
        ...             ...
        """
        df_or_ts = self._add_inflation()
        if real:
            df_or_ts = self._make_real_return_time_series(df_or_ts)
        df = self._make_df_if_series(df_or_ts)
        return helpers.Frame.get_rolling_fn(df, window=window, fn=helpers.Frame.get_cagr)

    def get_monthly_geometric_mean_return(self):
        """
        Calculate monthly geometric mean return for the portfolio.

        The geometric mean return is the constant periodic return that would produce the same
        final value as a sequence of varying periodic returns when compounded over time.

        Returns
        -------
        float
            Monthly geometric mean return value.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US'], weights=[0.6, 0.4])
        >>> pf.get_monthly_geometric_mean_return()
        0.005321
        """
        return (self.ror + 1.0).prod() ** (1 / self.ror.shape[0]) - 1.0

    def get_cumulative_return(self, period: Union[str, int, None] = None, real: bool = False) -> pd.Series:
        """
        Calculate cumulative return over a given trailing period for the portfolio.

        The cumulative return is the total change in the portfolio price during the investment period.

        Inflation adjusted cumulative returns (real cumulative returns) are shown with `real=True` option.
        Annual inflation data is calculated for the same period if `inflation=True` in the `Portfolio`.

        Parameters
        ----------
        period : str or int or None, default None
            Trailing period in years.
            None - full time cumulative return.
            'YTD' - (Year To Date) period of time beginning the first day of the calendar year up to the last month.
        real : bool, default False
            Cumulative return is adjusted for inflation (real cumulative return) if True.
            Portfolio should be initiated with `Inflation=True` for real cumulative return.

        Returns
        -------
        Series
            Cumulative rate of return values for portfolio and cumulative inflation (if inflation=True in Portfolio).

        Examples
        --------
        >>> pf = ok.Portfolio(['BTC-USD.CC', 'LTC-USD.CC'], weights=[.8, .2], last_date='2021-03')
        >>> pf.get_cumulative_return(period=2, real=True)
        portfolio_6232.PF    9.39381
        dtype: float64
        """
        ts = self._add_inflation()
        df = self._make_df_if_series(ts)
        dt0 = self.last_date

        if period is None:
            dt = self.first_date
        elif str(period).lower() == "ytd":
            year = dt0.year
            dt = str(year)
        else:
            self._validate_period(period)
            dt = helpers.Date.subtract_years(dt0, period)

        cr = helpers.Frame.get_cumulative_return(df[dt:])
        if real:
            if not hasattr(self, "inflation"):
                raise ValueError(
                    "Real cumulative return is not defined (no inflation information is available)."
                    "Set inflation=True in Portfolio to calculate it."
                )
            cumulative_inflation = helpers.Frame.get_cumulative_return(self.inflation_ts[dt:])
            cr = (1.0 + cr) / (1.0 + cumulative_inflation) - 1.0
            cr.drop(self.inflation, inplace=True)
        return cr

    def get_rolling_cumulative_return(self, window: int = 12, real: bool = False) -> pd.DataFrame:
        """
        Calculate rolling cumulative return.

        The cumulative return is the total change in the portfolio price.

        Parameters
        ----------
        window : int, default 12
            Size of the moving window in months.
        real : bool, default False
            Cumulative return is adjusted for inflation (real cumulative return) if True.
            Portfolio should be initiated with `Inflation=True` for real cumulative return.

        Returns
        -------
        DataFrame
            Time series of rolling cumulative return and inflation (optional).

        Examples
        --------
        >>> pf = ok.Portfolio(
        ...     ['SPY.US', 'AGG.US', 'GLD.US'],
        ...     weights=[.6, .35, .05],
        ...     rebalancing_strategy=ok.Rebalance(period="year"),
        ... )
        >>> pf.get_rolling_cumulative_return(window=24, real=True)
                 portfolio_9012.PF
        2006-11           0.125728
        2006-12           0.104348
        2007-01           0.129601
        2007-02           0.110680
        2007-03           0.132610
                            ...
        2021-03           0.263755
        2021-04           0.275474
        2021-05           0.322736
        2021-06           0.264963
        2021-07           0.273801
        [177 rows x 1 columns]
        """
        ts = self._add_inflation()
        if real:
            ts = self._make_real_return_time_series(ts)
        df = self._make_df_if_series(ts)
        return helpers.Frame.get_rolling_fn(
            df,
            window=window,
            fn=helpers.Frame.get_cumulative_return,
            window_below_year=True,
        )

    @property
    def assets_close_monthly(self) -> pd.DataFrame:
        """
        Show assets monthly close time series adjusted to the base currency.

        Returns
        -------
        DataFrame
            Assets monthly close time series adjusted to the base currency.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD')
        >>> pf.assets_close_monthly.plot()
        >>> plt.show()
        """
        series_list = []  # Collect all series to concatenate once at the end
        for x in self.asset_obj_dict.values():
            series = (
                x.close_monthly
                if x.currency == self.currency
                else self._adjust_price_to_currency_monthly(x.close_monthly, x.currency)
            )
            series = series.rename(x.symbol)
            series_list.append(series)
        if len(series_list) == 1:
            assets_close_monthly = series_list[0].to_frame()
        else:
            assets_close_monthly = pd.concat(series_list, axis=1, join="inner", copy="false")
        assets_close_monthly = assets_close_monthly[self.first_date : self.last_date]
        return assets_close_monthly

    @property
    def close_monthly(self) -> pd.Series:
        """
        Portfolio size monthly time series.

        Portfolio size is shown in base currency units. It is similar to the close value of an asset.
        Initial portfolio value is equal to 1000 units of base currency.

        Returns
        -------
        pd.Series
            Monthly portfolio size time series.

        Notes
        -----
        'close_mothly' shows the same output as the 'wealth_index'.
        This property is required as Portfolio must have the same attributes as an Asset.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD')
        >>> pf.close_monthly.plot()
        >>> plt.show()
        """
        return self.wealth_index.iloc[:, 0]

    @property
    def number_of_securities(self) -> pd.DataFrame:
        """
        Calculate the number of securities monthly time series for the portfolio assets.

        The number of securities in the Portfolio is changing over time as the dividends are reinvested.
        Portfolio rebalancing also affects the number of securities.

        Initial number of securities depends on the portfolio size in base currency (1000 units).

        Returns
        -------
        DataFrame
            Number of securities monthly time series for the portfolio assets.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD', last_date='07-2021')
        >>> pf.number_of_securities
                   SPY.US     BND.US
        Date
        2007-05  3.261153   6.687174
        2007-06  3.337216   6.758447
        2007-07  3.407015   6.643519
        2007-08  3.410268   6.663862
        2007-09  3.372630   6.798730
                   ...        ...
        2021-03  3.273521  15.313911
        2021-04  3.204779  15.685601
        2021-05  3.196768  15.749127
        2021-06  3.186124  15.879056
        2021-07  3.166335  16.003569
        [171 rows x 2 columns]
        """
        df = self.weights_ts.mul(self.wealth_index.iloc[:, 0], axis=0).div(self.assets_close_monthly, axis=0)
        return helpers.Frame.change_columns_order(df, self.symbols)

    @property
    def dividends(self) -> pd.Series:
        """
        Calculate portfolio dividends monthly time series.

        Portfolio dividends are obtained by summing asset dividends adjusted to the base currency.
        Dividends size depends on the portfolio value and number of securities.

        Returns
        -------
        Series
            Portfolio dividends monthly time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD', last_date='07-2021')
        >>> pf.dividends
        2007-05    0.849271
        2007-06    3.928855
        2007-07    1.551262
        2007-08    2.023148
        2007-09    4.423416
                     ...
        2021-03    6.155337
        2021-04    3.019478
        2021-05    2.056836
        2021-06    6.519521
        2021-07    2.114071
        Freq: M, Name: portfolio_2951.PF, Length: 171, dtype: float64
        """
        s = (self._get_assets_dividends() * self.number_of_securities).sum(axis=1)
        s.rename(self.symbol, inplace=True)
        return s

    @property
    def dividend_yield(self) -> pd.Series:
        """
        Calculate last twelve months (LTM) dividend yield time series for the portfolio. Time series has monthly values.

        Portfolio dividend yield is a weighted sum of the assets dividend yields (adjusted to
        the portfolio base currency).

        For an asset LTM dividend yield is the sum trailing twelve months of common dividends per share divided by
        the current price per share.

        Returns
        -------
        Series
            Portfolio LTM dividend yield monthly time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['T.US', 'XOM.US'], weights=[0.8, 0.2], first_date='2010-01', last_date='2021-01', ccy='USD')
        >>> pf.dividend_yield
        2010-01    0.013249
        2010-02    0.014835
        2010-03    0.014257
                     ...
        2020-11    0.076132
        2020-12    0.074743
        2021-01    0.073643
        Freq: M, Name: portfolio_8836.PF, Length: 133, dtype: float64

        >>> import matplotlib.pyplot as plt
        >>> pf.dividend_yield.plot()
        >>> plt.show()
        """
        df = self._assets_dividend_yield @ self.weights_ts.T
        div_yield_series = pd.Series(np.diag(df), index=df.index)  # faster than df1.mul(df2).sum(axis=1)
        div_yield_series.rename(self.symbol, inplace=True)
        return div_yield_series

    @property
    def dividends_annual(self) -> pd.DataFrame:
        """
        Return calendar year dividends sum time series for each asset.

        Returns
        -------
        DataFrame
            Annual dividends time series for each asset.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD', last_date='07-2021')
        >>> pf.dividends_annual.plot(kind='bar')
        >>> plt.show()
        """
        return self._get_assets_dividends().resample("Y").sum()

    @property
    def dividend_yield_annual(self):
        """
        Calculate last twelve months (LTM) dividend yield annual time series.

        Time series is based on the dividend yield for the end of calendar year.

        LTM dividend yield is the sum trailing twelve months of common dividends per share divided by
        the current price per share.

        All yields are calculated in the asset list base currency after adjusting the dividends and price time series.
        Forecasted (future) dividends are removed.

        Returns
        -------
        DataFrame
            Time series of LTM dividend yield for each asset.

        See Also
        --------
        dividend_yield : Dividend yield time series.
        dividends_annual : Calendar year dividends time series.
        dividend_paying_years : Number of years of consecutive dividend payments.
        dividend_growing_years : Number of years when the annual dividend was growing.
        get_dividend_mean_yield : Arithmetic mean for annual dividend yield.
        get_dividend_mean_growth_rate : Geometric mean of annual dividends growth rate.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD', last_date='07-2021')
        >>> pf.dividend_yield_annual.plot(kind='bar')
        >>> plt.show()
        """
        return self._assets_dividend_yield.resample(rule="Y").last()

    @property
    def assets_dividend_yield(self):
        """
        Calculate last twelve months (LTM) dividend yield time series (monthly) for each asset.

        LTM dividend yield is the sum trailing twelve months of common dividends per share divided by
        the current price per share.

        All yields are calculated in the asset list base currency after adjusting the dividends and price time series.
        Forecasted (future) dividends are removed.
        Zero value time series are created for assets without dividends.

        Returns
        -------
        DataFrame
            Monthly time series of LTM dividend yield for each asset.

        See Also
        --------
        dividend_yield_annual : Calendar year dividend yield time series.
        dividends_annual : Calendar year dividends time series.
        dividend_paying_years : Number of years of consecutive dividend payments.
        dividend_growing_years : Number of years when the annual dividend was growing.
        get_dividend_mean_yield : Arithmetic mean for annual dividend yield.
        get_dividend_mean_growth_rate : Geometric mean of annual dividends growth rate.

        Examples
        --------
        >>> x = ok.AssetList(['T.US', 'XOM.US'], first_date='1984-01', last_date='1994-12')
        >>> x.dividend_yield
                   T.US    XOM.US
        1984-01  0.000000  0.000000
        1984-02  0.000000  0.002597
        1984-03  0.002038  0.002589
        1984-04  0.001961  0.002346
                   ...       ...
        1994-09  0.018165  0.012522
        1994-10  0.018651  0.011451
        1994-11  0.018876  0.012050
        1994-12  0.019344  0.011975
        [132 rows x 2 columns]
        """
        return super()._assets_dividend_yield

    @property
    def real_mean_return(self) -> float:
        """
        Calculate annualized real mean return (arithmetic mean) for the rate of return time series.

        Real rate of return is adjusted for inflation. Real return is defined if
        there is an `inflation=True` option in Portfolio.

        Returns
        -------
        float
            Annualized value of the mean for the real rate of return time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.real_mean_return
        0.3088967455111862
        """
        if not hasattr(self, "inflation"):
            raise ValueError("Real Return is not defined. Set inflation=True to calculate.")
        infl_mean = helpers.Float.annualize_return(self.inflation_ts.mean())
        ror_mean = helpers.Float.annualize_return(self.ror.mean())
        return (1.0 + ror_mean) / (1.0 + infl_mean) - 1.0

    @property
    def risk_monthly(self) -> pd.Series:
        """
        Calculate monthly risk expanding time series for Portfolio.

        Monthly risk of portfolio is a standard deviation of the rate of return time series.
        Standard deviation (sigma) is normalized by N-1.

        Returns
        -------
        Series
            Standard deviation of the monthly return expanding time series.

        See Also
        --------
        risk_annual : Calculate annualized risks.
        semideviation_monthly : Calculate semideviation monthly values.
        semideviation_annual : Calculate semideviation annualized values.
        get_var_historic : Calculate historic Value at Risk (VaR).
        get_cvar_historic : Calculate historic Conditional Value at Risk (CVaR).
        drawdowns : Calculate drawdowns.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.risk_monthly
        date
        1986-05    0.020117
        1986-06    0.122032
        1986-07    0.130113
        1986-08    0.116642
                     ...
        2023-08    0.092875
        2023-09    0.092861
        2023-10    0.092759
        2023-11    0.092763
        2023-12    0.092665
        Freq: M, Name: portfolio_1094.PF, Length: 453, dtype: float64
        """
        return self.ror.expanding().std().iloc[1:]

    @property
    def risk_annual(self) -> pd.Series:
        """
        Calculate annualized risk expanding time series for portfolio.

        Risk is a standard deviation of the rate of return.

        Annualized risk is calculated for rate of retirun time series for the sample from 'first_date' to
        'last_date'.

        Returns
        -------
        Series
            Annualized standard deviation of the monthly return expanding time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.risk_annual
        date
        1986-05    0.285175
        1986-06    0.890909
        1986-07    0.616876
        1986-08    0.632270
        1986-09    0.509642
                     ...
        2023-08    0.428297
        2023-09    0.427350
        2023-10    0.426961
        2023-11    0.427930
        """
        risk_ts = self.ror.expanding().std()
        mean_return_ts = self.ror.expanding().mean()
        return helpers.Float.annualize_risk(risk_ts, mean_return_ts).iloc[1:]

    @property
    def semideviation_monthly(self) -> float:
        """
        Calculate semi-deviation monthly value for portfolio rate of return time series.

        Semi-deviation (Downside risk) is the risk of the return being below the expected return.

        Returns
        -------
        float
            Semi-deviation monthly value for portfolio rate of return time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.semideviation_monthly
        0.05601433676604449
        """
        return helpers.Frame.get_semideviation(self.ror)

    @property
    def semideviation_annual(self) -> float:
        """
        Return semideviation annualized value for portfolio rate of return time series.

        Semi-deviation (Downside risk) is the risk of the return being below the expected return.

        Returns
        -------
        float
            Annualized semi-deviation monthly value for portfolio rate of return time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.semideviation_annual
        0.1940393544621248
        """
        return helpers.Frame.get_semideviation(self.ror) * 12**0.5

    def get_var_historic(self, time_frame: int = 12, level=1) -> float:
        """
        Calculate historic Value at Risk (VaR) for the portfolio.

        The VaR calculates the potential loss of an investment with a given time frame and confidence level.
        Loss is a positive number (expressed in cumulative return).
        If VaR is negative there are expected gains at this confidence level.

        Parameters
        ----------
        time_frame : int, default 12
            Time frame for VAR. Default is 12 months.
        level : int, default 1
            Confidence level in percents. Default value is 1%.

        Returns
        -------
        float
            Historic Value at Risk (VaR) value for the portfolio.

        Examples
        --------
        >>> x = ok.Portfolio(['SP500TR.INDX', 'SP500BDT.INDX'], last_date='2021-01')
        >>> x.get_var_historic(time_frame=12, level=1)
        0.24030006476701732
        """
        # remove inflation column from rolling return
        df = self.get_rolling_cumulative_return(window=time_frame).loc[:, [self.symbol]]
        return helpers.Frame.get_var_historic(df, level).iloc[0]

    def get_cvar_historic(self, time_frame: int = 12, level=1) -> float:
        """
        Calculate historic Conditional Value at Risk (CVAR, expected shortfall) for the portfolio.

        CVaR is the average loss over a specified time period of unlikely scenarios beyond the confidence level.
        Loss is a positive number (expressed in cumulative return).
        If CVaR is negative there are expected gains at this confidence level.

        Parameters
        ----------
        time_frame : int, default 12
            Time period size in months.
        level : int, default 1
            Confidence level in percents to calculate the VaR. Default value is 1% (1% quantile).

        Returns
        -------
        float
            Historic Conditional Value at Risk (CVAR, expected shortfall) value for the portfolio.

        Examples
        --------
        >>> x = ok.Portfolio(['USDEUR.FX', 'BTC-USD.CC'], last_date='2021-01')
        >>> x.get_cvar_historic(time_frame=2, level=1)
        0.3566909250442616
        """
        # remove inflation column form rolling return
        df = self.get_rolling_cumulative_return(window=time_frame).loc[:, [self.symbol]]
        return helpers.Frame.get_cvar_historic(df, level).iloc[0]

    @property
    def drawdowns(self) -> pd.Series:
        """
        Calculate drawdowns time series for the portfolio.

        The drawdown is the percent decline from a previous peak in wealth index.

        Returns
        -------
        Series
            Drawdowns time series for the portfolio
        """
        return helpers.Frame.get_drawdowns(self.ror)

    @property
    def recovery_period(self) -> pd.Series:
        """
        Get recovery period time series for the portfolio value.

        The recovery period (drawdown duration) is the number of months to reach the value of the last maximum.

        Returns
        -------
        pd.Series
            Recovery period time series for the portfolio value

        Notes
        -----
        The largest recovery period does not necessary correspond to the max drawdown.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US'], weights=[0.5, 0.5])
        >>> pf.recovery_period.nlargest()
        date
        2010-10    35
        2004-10     7
        2012-01     7
        2019-03     6
        2018-07     5
        Freq: M, Name: portfolio_5724.PF, dtype: int32

        See Also
        --------
        drawdowns : Calculate drawdowns time series.
        """
        w_index = self.wealth_index_with_assets[self.symbol]
        cummax = w_index.cummax()
        s = cummax.pct_change()[1:]
        s1 = s.where(s == 0).notnull().astype(int)
        s1_1 = s.where(s == 0).isnull().astype(int).cumsum()
        s2 = s1.groupby(s1_1).cumsum()
        return s2[s2.shift(-1) < s2]

    def describe(self, years: Tuple[int] = (1, 5, 10)) -> pd.DataFrame:
        """
        Generate descriptive statistics for the portfolio.

        Statistics includes:

        - YTD (Year To date) compound return
        - CAGR for a given list of periods and full available period
        - Annualized mean rate of return (full available period)
        - LTM Dividend yield - last twelve months dividend yield

        Risk metrics (full available period):

        - risk (standard deviation)
        - CVAR (timeframe is 1 year)
        - max drawdowns (and dates)

        Parameters
        ----------
        years : tuple of int, default (1, 5, 10)
            List of periods for CAGR statistics.

        Returns
        -------
        DataFrame
            Table of descriptive statistics for the portfolio.

        See Also
        --------
        get_cumulative_return : Calculate cumulative return.
        get_cagr : Calculate portfolio Compound Annual Growth Rate (CAGR).
        dividend_yield : Calculate dividend yield (LTM).
        risk_annual : Return annualized risks (standard deviation).
        get_cvar_historic : Calculate historic Conditional Value at Risk (CVAR, expected shortfall).
        drawdowns : Calculate drawdowns.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD', last_date='07-2021')
        >>> pf.describe(years=[2, 5, 7])  # 'years' customizes the timeframe for the CAGR
                    property              period portfolio_2951.PF  inflation
        0    compound return                 YTD          0.084098   0.048154
        1               CAGR             2 years          0.141465   0.031566
        2               CAGR             5 years          0.102494   0.025582
        3               CAGR             7 years          0.091694   0.019656
        4               CAGR  14 years, 3 months          0.074305   0.019724
        5     Dividend yield                 LTM          0.016504        NaN
        6               Risk  14 years, 3 months          0.086103        NaN
        7               CVAR  14 years, 3 months          0.214207        NaN
        8       Max drawdown  14 years, 3 months         -0.266915        NaN
        9  Max drawdown date  14 years, 3 months           2009-02        NaN
        """
        rows_list = []
        dt0 = self.last_date
        df = self._add_inflation()
        # YTD return
        ytd_return = self.get_cumulative_return(period="YTD")
        row = ytd_return.to_dict()
        row.update(period="YTD", property="compound return")
        rows_list.append(row)
        # CAGR for a list of periods
        if self.pl.years >= 1:
            for i in years:
                dt = helpers.Date.subtract_years(dt0, i)
                if dt >= self.first_date:
                    row = self.get_cagr(period=i).to_dict()
                else:
                    row = {x: None for x in df.columns} if hasattr(self, "inflation") else {self.symbol: None}
                row.update(period=f"{i} years", property="CAGR")
                rows_list.append(row)
            # CAGR for full period
            row = self.get_cagr(period=None).to_dict()
            row.update(
                period=self._pl_txt,
                property="CAGR",
            )
            rows_list.append(row)
            # Mean rate of return (arithmetic mean)
            value = self.mean_return_annual
            row = {self.symbol: value}
            row.update(
                period=self._pl_txt,
                property="Annualized mean return",
            )
            rows_list.append(row)
            # Dividend Yield
            value = self.dividend_yield.iloc[-1]
            row = {self.symbol: value}
            row.update(
                period="LTM",
                property="Dividend yield",
            )
            rows_list.append(row)
        # risk (standard deviation)
        row = {self.symbol: self.risk_annual.iloc[-1]}
        row.update(period=self._pl_txt, property="Risk")
        rows_list.append(row)
        # CVAR
        if self.pl.years >= 1:
            row = {self.symbol: self.get_cvar_historic(level=1)}
            row.update(
                period=self._pl_txt,
                property="CVAR (α=1)",
            )
            rows_list.append(row)
        # max drawdowns
        row = {self.symbol: self.drawdowns.min()}
        row.update(
            period=self._pl_txt,
            property="Max drawdown",
        )
        rows_list.append(row)
        # max drawdowns dates
        row = {self.symbol: self.drawdowns.idxmin()}
        row.update(
            period=self._pl_txt,
            property="Max drawdown date",
        )
        rows_list.append(row)
        # Create DataFrame from list of rows
        description = pd.DataFrame(rows_list)
        if hasattr(self, "inflation"):
            description.rename(columns={self.inflation: "inflation"}, inplace=True)
        description = helpers.Frame.change_columns_order(description, ["property", "period", self.symbol])
        return description

    @property
    def table(self) -> pd.DataFrame:
        """
        Return table with security name, ticker, weight for assets in the portfolio.

        Returns
        -------
        DataFrame
            Security name - ticker - weight table.

        Examples
        --------
        >>> pf = ok.Portfolio(["MSFT.US", "AAPL.US"])
        >>> pf.table
                        asset name   ticker  weights
        0  Microsoft Corporation  MSFT.US      0.5
        1              Apple Inc  AAPL.US      0.5
        """
        x = pd.DataFrame(
            data={
                "asset name": list(self.names.values()),
                "ticker": list(self.names.keys()),
            }
        )
        x["weights"] = self.weights
        return x

    # Historical percentiles
    def percentile_inverse_cagr(
        self,
        years: int = 1,
        score: float = 0,
    ) -> float:
        """
        Compute the percentile rank of a score (CAGR value).

        Percentile rank can be calculated for a historical distribution of CAGR.

        If percentile_inverse of, for example, 0% (CAGR value) is equal to 8% for 1 year time frame
        it means that 8% of the CAGR values in the distribution are negative in 1 year periods. Or in other words
        the probability of getting negative result after 1 year of investments is 8%.

        Parameters
        ----------
        years : int, default 1
            Period length (time frame) in years when CAGR is calculated.
        score : float, default 0
            Score that is compared to the elements in CAGR array.

        Returns
        -------
        float
            Percentile-position of score (0-100) relative to distribution.

        Examples
        --------
        >>> pf = ok.Portfolio(
        ...     ['SPY.US', 'AGG.US', 'GLD.US'],
        ...     weights=[.60, .35, .05],
        ...     rebalancing_strategy=ok.Rebalance(period="year"),
        ... )
        >>> pf.percentile_inverse_cagr(score=0, years=1)
        18.08
        The probability of getting negative result (score=0) in 1 year period for historical distribution.
        """
        cagr_distr = self.get_rolling_cagr(years * settings._MONTHS_PER_YEAR).loc[:, [self.symbol]].squeeze()
        return scipy.stats.percentileofscore(cagr_distr, score, kind="rank")

    def percentile_cagr(self, years: int, percentiles: List[int] = [10, 50, 90]) -> pd.DataFrame:
        """
        Calculate given percentiles for portfolio rolling CAGR distribution from the historical data.

        CAGR - Compound Annual Growth Rate.
        Each percentile is calculated for a period range from 1 year to 'years'.

        Parameters
        ----------
        years : int
            Max window size for rolling CAGR in the distribution in years.
            It should not exceed 1/2 of the portfolio history period length 'period_length'.
        percentiles : list of int, default [10, 50, 90]
            List of percentiles to be calculated.

        Returns
        -------
        DataFrame
            Table with percentiles values for each period from 1 to 'years'.

        Examples
        --------
        >>> pf = ok.Portfolio(
        ...     ['SPY.US', 'AGG.US', 'GLD.US'],
        ...     weights=[.60, .35, .05],
        ...     rebalancing_strategy=ok.Rebalance(period="none"),
        ... )
        >>> pf.percentile_cagr(years=5, percentiles=[1, 50, 99])
                     1         50        99
        years
        1     -0.231327  0.098693  0.295343
        2     -0.101689  0.091824  0.206471
        3     -0.036771  0.085428  0.157833
        4     -0.007674  0.085178  0.142195
        5      0.030933  0.082865  0.134496
        """
        period_range = range(1, years + 1)
        returns_dict = {}
        for percentile in percentiles:
            percentile_returns_list = [
                self.get_rolling_cagr(years * 12).loc[:, self.symbol].quantile(percentile / 100)
                for years in period_range
            ]
            returns_dict.update({percentile: percentile_returns_list})
        df = pd.DataFrame(returns_dict, index=list(period_range))
        df.index.rename("years", inplace=True)
        return df

    def get_sharpe_ratio(self, rf_return: float = 0) -> float:
        """
        Calculate Sharpe ratio.

        The Sharpe ratio is the average annual return in excess of the risk-free rate
        per unit of risk (annualized standard deviation).

        Risk-free rate should be taken according to the Portfolio base currency.

        Parameters
        ----------
        rf_return : float, default 0
            Risk-free rate of return.

        Returns
        -------
        float

        Examples
        --------
        >>> pf = ok.Portfolio(['VOO.US', 'BND.US'], weights=[0.40, 0.60])
        >>> pf.get_sharpe_ratio(rf_return=0.04)
        0.7412193684695373
        """
        return ratios.get_sharpe_ratio(
            pf_return=self.mean_return_annual,
            rf_return=rf_return,
            std_deviation=self.risk_annual.iloc[-1],
        )

    def get_sortino_ratio(self, t_return: float = 0) -> float:
        """
        Calculate Sortino ratio for the portfolio with specified target return.

        Sortion ratio measures the risk-adjusted return of portfolio. It is a modification of the Sharpe ratio
        but penalizes only those returns falling below a specified target rate of return, while
        the Sharpe ratio penalizes both upside and downside volatility equally.

        Parameters
        ----------
        t_return : float, default 0
            Traget rate of return.

        Returns
        -------
        float

        Examples
        --------
        >>> pf = ok.Portfolio(['VOO.US', 'BND.US'], last_date='2021-12')
        >>> pf.get_sortino_ratio(t_return=0.02)
        1.4377728903230174
        """
        semideviation = helpers.Frame.get_below_target_semideviation(ror=self.ror, t_return=t_return) * 12**0.5
        return ratios.get_sortino_ratio(
            pf_return=self.mean_return_annual,
            t_return=t_return,
            semi_deviation=semideviation,
        )

    @property
    def diversification_ratio(self) -> float:
        """
        Calculate Diversification Ratio for the portfolio.

        The Diversification Ratio is the ratio of the weighted average of assets risks divided by the portfolio risk.
        In this case risk is the annuilized standatd deviation for the rate of return .

        Returns
        -------
        float

        Examples
        --------
        >>> pf = ok.Portfolio(['VOO.US', 'BND.US'], weights=[0.7, 0.3], last_date='2021-12')
        >>> pf.diversification_ratio
        1.1264305597257505
        """
        assets_risk = self.assets_ror.std()
        assets_mean_return = self.assets_ror.mean()
        assets_annualized_risk = helpers.Float.annualize_risk(assets_risk, assets_mean_return)
        weights = np.asarray(self.weights)
        sigma_weighted_sum = weights.T @ assets_annualized_risk
        return sigma_weighted_sum / self.risk_annual.iloc[-1]

    @property
    def okamaio_link(self) -> str:
        """
        URL link to portfolio at okama.io.

        Portfolio with the same tickers, weights and other properties at okama.io financial widgets.

        Returns
        -------
        str
            URL link to portfolio at okama.io.

        Examples
        --------
        >>> pf = ok.Portfolio(
        ...     ['SPY.US', 'AGG.US'],
        ...     weights=[.60, .40],
        ...     rebalancing_strategy=ok.Rebalance(period="year", abs_deviation=0.05),
        ... )
        >>> pf.okamaio_link
        'https://okama.io/portfolio?tickers=SPY.US,AGG.US&weights=60.0,40.0&ccy=USD&first_date=2003-10-01&last_date=2024-08-01&rebal=year&rebalancing_period=year&rebalancing_abs_deviation=0.05&symbol=portfolio_6323.PF'
        """
        weights_percent = [w * 100 for w in self.weights]
        query_params = {
            "tickers": ",".join(str(symbol) for symbol in self.symbols),
            "weights": ",".join(str(weight) for weight in weights_percent),
            "ccy": self.currency,
            "first_date": self.first_date.strftime("%Y-%m-%d"),
            "last_date": self.last_date.strftime("%Y-%m-%d"),
            "rebal": self.rebalancing_strategy.period,
            "rebalancing_period": self.rebalancing_strategy.period,
            "symbol": self.symbol,
        }
        if self.rebalancing_strategy.abs_deviation is not None:
            query_params["rebalancing_abs_deviation"] = self.rebalancing_strategy.abs_deviation
        if self.rebalancing_strategy.rel_deviation is not None:
            query_params["rebalancing_rel_deviation"] = self.rebalancing_strategy.rel_deviation
        return f"https://okama.io/portfolio?{urlencode(query_params, safe=',')}"

    def _clear_cf_cache(self):
        self._monte_carlo_wealth = pd.DataFrame()
        self._wealth_index = pd.DataFrame()
        self._cash_flow_fv = pd.DataFrame()



from typing import Optional, Tuple, Union
from dataclasses import dataclass

import numpy as np
import pandas as pd

from okama import settings
from okama.common.validators import validate_real


@dataclass(frozen=True)
class Result:
    """
    The result of rebalancing portfolio.

    Attributes
    ----------
    portfolio_wealth_index : Series
        Monthly time series for rebalanced portfolio wealth index.

    assets_wealth_indexes : DataFrame
        The absolute amount of withdrawal size (the best solution found).

    events : Series
        The relative amount of withdrawal size (the best solution found).
    """

    portfolio_wealth_index: pd.Series
    assets_wealth_indexes: pd.DataFrame
    events: pd.Series


class Rebalance:
    """
    Rebalancing strategy for an investment portfolio.

    Rebalancing is the process by which an investor restores their portfolio to its target allocation
    by selling and buying assets. After rebalancing all the assets have original (target) weights.

    Attributes
    ----------
    period : {'none', 'month', 'quarter', 'half-year', 'year'}
        rebalancing period (rebalancing frequency) is predetermined time intervals when
        the investor rebalances the portfolio.
        If 'none' assets weights are not rebalanced.

    abs_deviation : float, optional
        The absolute deviation allowed for the assets weights in the portfolio.
        It is defined for an asset weight as an absolut value of: actual_weight - target_weight.
        It must be more than 0 (0%).
        Max value is 1 (100%).
        if None the parameter is ignored.

    rel_deviation : float, optional
        The relative deviation allowed for the assets weights in the portfolio.
        It is defined for an asset weight as an absolut value of: actual_weight / target_weight - 1.
        it must be positive.
        if None the parameter is ignored.
    """

    def __init__(
        self, period: str = "year", abs_deviation: Optional[float] = None, rel_deviation: Optional[float] = None
    ):
        self.abs_deviation = abs_deviation
        self.rel_deviation = rel_deviation
        self.period = period

    def __str__(self):
        dic = {
            "period": self.period,
            "abs_deviation": self.abs_deviation,
            "rel_deviation": self.rel_deviation,
        }
        return repr(pd.Series(dic))

    def __repr__(self):
        dic = {
            "period": self.period,
            "abs_deviation": self.abs_deviation,
            "rel_deviation": self.rel_deviation,
        }
        return repr(pd.Series(dic))

    def _validate_condition(self):
        if self.period not in settings.frequency_mapping.keys():
            raise ValueError(f"rebalancing_period must be in {list(settings.frequency_mapping.keys())}")
        if self.abs_deviation:
            validate_real(arg_name="abs_deviation", arg_value=self.abs_deviation)
            if self.abs_deviation <= 0:
                raise ValueError("Absolute deviation must be positive.")
            if self.abs_deviation > 1:
                raise ValueError("Absolute deviation must be less or equal to 1.")
        if self.rel_deviation:
            validate_real(arg_name="rel_deviation", arg_value=self.rel_deviation)
            if self.rel_deviation <= 0:
                raise ValueError("Relative deviation must be positive.")

    @property
    def period(self) -> str:
        """
        The rebalancing period for the investment portfolio.

        It can be: 'none', 'month', 'quarter', 'half-year', 'year'.
        """
        return self._period

    @period.setter
    def period(self, value: str):
        self._period = value
        self._validate_condition()
        self._pandas_frequency = settings.frequency_mapping[self.period]
        self._pandas_frequency_grouper = settings.grouper_frequency_mapping[self.period]

    def wealth_ts(
        self, target_weights: list, ror: pd.DataFrame, calculate_assets_wealth_indexes: bool = False
    ) -> Result:
        """
        Calculate wealth index time series of rebalanced portfolio given returns time series of the assets.

        Optionally calculate also ASSETS wealth indexes time series inside rebalanced portfolio.

        Parameters
        ----------
        target_weights : list of float
            The target weights for assets in the portfolio.

        ror : pd.DataFrame
            Assets rate of return monthly time series.

        calculate_assets_wealth_indexes : bool, optional
            Whether or not to calculate assets wealth indexes after rebalancing.
            When 'True' works slower.

        Returns
        -------
        Result
            Result of rebalancing investment portfolio.
        """
        if isinstance(ror, pd.Series):
            ror = ror.to_frame()
        if len(target_weights) != len(ror.columns):
            raise ValueError("The dimension of target_weights and the number of columns in ror must be equal")
        initial_inv = 1000
        first_date = ror.index[0]
        first_wealth_index_date = first_date - 1
        portfolio_wealth_index = pd.Series(dtype="float64")
        assets_wealth_indexes = pd.DataFrame(columns=ror.columns, dtype="float64")
        events_ts = pd.Series(dtype="float64")
        target_weights_np = np.asarray(target_weights)
        if self.period == "none" and self.abs_deviation is None and self.rel_deviation is None:  # No rebalancing
            initial_allocation = target_weights_np * initial_inv
            assets_wealth_indexes_local = initial_allocation * (1 + ror).cumprod()
            portfolio_wealth_index = assets_wealth_indexes_local.sum(axis=1)
            if calculate_assets_wealth_indexes:
                assets_wealth_indexes = assets_wealth_indexes_local
        elif self.period == "none":  # No calendar rebalancing
            portfolio_wealth_index, assets_wealth_indexes, events_ts = self._rebalance_by_condition(
                ror, target_weights, initial_inv, calculate_assets_wealth_indexes
            )
        else:  # Calendar rebalancing
            rebalancing_by_condition_needed = self.abs_deviation or self.rel_deviation
            rebalancing_condition = False
            # accumulate chunks and concatenate once after the loop
            pw_chunks: list[pd.Series] = []
            awi_chunks: list[pd.DataFrame] = [] if calculate_assets_wealth_indexes else None
            for n, x in enumerate(ror.resample(rule=self._pandas_frequency, convention="start")):
                df = x[1]  # select ror part of the grouped data
                if n == 0:
                    initial_allocation = target_weights_np * initial_inv
                else:
                    if (rebalancing_by_condition_needed and rebalancing_condition) or (
                        not rebalancing_by_condition_needed
                    ):
                        # rebalancing
                        initial_allocation = target_weights_np * end_period_balance
                        date = x[0]
                        if rebalancing_by_condition_needed:
                            events_ts[date.asfreq("M", how="start") - 1] = "abs" if condition_abs else "rel"
                        else:
                            events_ts[date.asfreq("M", how="start") - 1] = "calendar"
                    elif rebalancing_by_condition_needed and not rebalancing_condition:
                        # skip rebalancing
                        initial_allocation = end_period_weights * end_period_balance
                assets_wealth_indexes_local = initial_allocation * (1 + df).cumprod()
                if calculate_assets_wealth_indexes:
                    # collect asset wealth index chunks for later concatenation
                    awi_chunks.append(assets_wealth_indexes_local.copy())
                wealth_index_local = assets_wealth_indexes_local.sum(axis=1)
                # collect portfolio wealth index chunks for later concatenation
                pw_chunks.append(wealth_index_local.copy())
                # use local last value instead of relying on concatenated series
                end_period_balance = wealth_index_local.iloc[-1]
                if rebalancing_by_condition_needed:
                    end_period_weights = assets_wealth_indexes_local.iloc[-1].divide(
                        wealth_index_local.iloc[-1], axis=0
                    )
                    rebalancing_condition, condition_abs = self._check_if_rebalancing_required(
                        assets_wealth_indexes_local, wealth_index_local, target_weights
                    )
            # perform single concatenation outside the loop
            portfolio_wealth_index = pd.concat(pw_chunks, sort=False) if pw_chunks else portfolio_wealth_index
            if calculate_assets_wealth_indexes and awi_chunks:
                assets_wealth_indexes = pd.concat(awi_chunks, verify_integrity=True, sort=False)
        # set value for the first date
        portfolio_wealth_index.loc[first_wealth_index_date] = initial_inv
        portfolio_wealth_index.sort_index(ascending=True, inplace=True)
        if calculate_assets_wealth_indexes:
            assets_wealth_indexes.loc[first_wealth_index_date] = target_weights_np * initial_inv
            assets_wealth_indexes.sort_index(ascending=True, inplace=True)
        return Result(
            portfolio_wealth_index=portfolio_wealth_index, assets_wealth_indexes=assets_wealth_indexes, events=events_ts
        )

    def _rebalance_by_condition(
        self, ror, target_weights, initial_inv, calculate_assets_wealth_indexes: bool = False
    ) -> Tuple[pd.Series, pd.DataFrame, pd.Series]:
        target_weights_np = np.asarray(target_weights)
        portfolio_wealth_index = pd.Series(dtype="float64")
        assets_wealth_indexes_local = pd.DataFrame(columns=ror.columns, dtype="float64")
        assets_wealth_indexes = pd.DataFrame(columns=ror.columns, dtype="float64")
        events_ts = pd.Series(dtype="float64")
        awi_rows = [] if calculate_assets_wealth_indexes else None
        for n, row in enumerate(ror.itertuples()):
            date = row[0]
            r = pd.Series(row[1:], index=ror.columns, name=date)
            if n == 0:
                initial_allocation = target_weights_np * initial_inv  # initial rebalancing
                assets_wealth_indexes_local = initial_allocation * (1 + r)
            else:
                if rebalancing_condition:
                    assets_wealth_indexes_local = target_weights_np * assets_wealth_indexes_local.sum()
                    events_ts[date - 1] = "abs" if condition_abs else "rel"  # set previous month as its EOD data
                assets_wealth_indexes_local *= 1 + r
                assets_wealth_indexes_local.rename(date, inplace=True)
            if calculate_assets_wealth_indexes:
                # collect rows for a single concat after the loop
                awi_rows.append(pd.DataFrame(assets_wealth_indexes_local).T)
            portfolio_wealth_index_local = assets_wealth_indexes_local.sum()
            portfolio_wealth_index[date] = portfolio_wealth_index_local
            # Check if rebalancing required
            rebalancing_condition, condition_abs = self._check_if_rebalancing_required(
                assets_wealth_indexes_local, portfolio_wealth_index_local, target_weights
            )
        # perform single concatenation outside the loop
        if calculate_assets_wealth_indexes and awi_rows:
            assets_wealth_indexes = pd.concat(awi_rows, sort=False)
        return portfolio_wealth_index, assets_wealth_indexes, events_ts

    def _check_if_rebalancing_required(
        self,
        assets_wealth_indexes_local,
        portfolio_wealth_index_local,
        target_weights,
    ) -> Tuple[bool, bool]:
        try:
            # DataFrame
            weights = assets_wealth_indexes_local.iloc[-1].divide(portfolio_wealth_index_local.iloc[-1], axis=0)
            target_weights_s = pd.Series(target_weights, index=assets_wealth_indexes_local.columns)
        except AttributeError:
            # Series
            weights = assets_wealth_indexes_local.divide(portfolio_wealth_index_local, axis=0)
            target_weights_s = pd.Series(target_weights, index=assets_wealth_indexes_local.index)
        weights_difference_abs = weights - target_weights_s
        weights_difference_abs = weights_difference_abs.abs()
        weights_difference_rel = weights.divide(target_weights_s, axis=0) - 1
        weights_difference_rel = weights_difference_rel.abs()
        # Ensure Python bools for downstream identity checks in tests
        condition_abs_np = (
            (weights_difference_abs > self.abs_deviation).any() if self.abs_deviation is not None else False
        )
        condition_rel_np = (
            (weights_difference_rel > self.rel_deviation).any() if self.rel_deviation is not None else False
        )
        condition_abs = bool(condition_abs_np)
        condition_rel = bool(condition_rel_np)
        # Determined at the end, as it is not needed during the first run.
        rebalancing_condition = bool(condition_abs or condition_rel)
        return rebalancing_condition, condition_abs

    def assets_weights_ts(self, target_weights: list, ror: pd.DataFrame) -> pd.DataFrame:
        """
        Calculate assets weights monthly time series for rebalanced portfolio.

        Parameters
        ----------
        target_weights : list of float
            The target weights for assets in the portfolio.

        ror : pd.DataFrame
            Assets rate of return monthly time series.

        Returns
        -------
        pd.DataFrame
            Assets weights monthly time series.
        """
        reb = self.wealth_ts(target_weights=target_weights, ror=ror, calculate_assets_wealth_indexes=True)
        return reb.assets_wealth_indexes.divide(reb.portfolio_wealth_index, axis=0)

    def return_ror_ts(self, target_weights: Union[list, np.ndarray], ror: pd.DataFrame) -> pd.Series:
        """
        Return monthly rate of return time series of rebalanced portfolio given returns time series of the assets.

        Parameters
        ----------
        target_weights : list of float
            The target weights for assets in the portfolio.

        ror : pd.DataFrame
            Assets rate of return monthly time series.

        Returns
        -------
        pd.Series
            The monthly rate of return time series of rebalanced portfolio.
        """
        # Calculate wealth index using wealth_ts (which includes an initial pre-first-period point)
        wealth_index = self.wealth_ts(target_weights, ror).portfolio_wealth_index
        # Align behavior with manual baseline: start pct_change from the first input ror date,
        # so the first period in ror.index will be NaN and then dropped.
        first_ror_date = ror.index[0]
        wealth_index_aligned = wealth_index.loc[wealth_index.index >= first_ror_date]
        portfolio_ror = wealth_index_aligned.pct_change()
        portfolio_ror.dropna(inplace=True)
        return portfolio_ror

    def wealth_ts_ef(self, weights: list, ror: pd.DataFrame) -> pd.Series:
        """
        Calculate wealth index time series of rebalanced portfolio given returns time series of the assets.

        Simple version without conditional rebalancing (abs_deviation and rel_deviation are not used).
        This method is used for efficient frontier optimization.
        Default rebalancing period is a Year (end of year).
        For not rebalanced portfolio set period to 'none'.

        Parameters
        ----------
        weights : list of float
            The target weights for assets in the portfolio.

        ror : pd.DataFrame
            Assets rate of return monthly time series.

        Returns
        -------
        pd.Series
            Wealth index time series.
        """
        initial_inv = 1000
        weights_array = np.asarray(weights)

        if self.period == "none":  # Not rebalanced portfolio
            inv_period_spread = weights_array * initial_inv
            assets_wealth_indexes = inv_period_spread * (1 + ror).cumprod()
            wealth_index = assets_wealth_indexes.sum(axis=1)
        else:
            # Fully vectorized approach without loops
            period_grouper = pd.Grouper(freq=self._pandas_frequency_grouper, convention="start")
            period_ids = ror.groupby(period_grouper).ngroup()

            growth_factors = 1 + ror
            cumulative_growth_within_period = growth_factors.groupby(period_ids).cumprod()

            wealth_per_asset_normalized = weights_array * cumulative_growth_within_period
            portfolio_wealth_normalized = wealth_per_asset_normalized.sum(axis=1)

            # Calculate ending wealth for each period (growth factor for that period)
            period_ends = portfolio_wealth_normalized.groupby(period_ids).last()

            # Calculate cumulative multiplier for the START of each period
            period_multipliers = period_ends.shift(1).fillna(1.0).cumprod()

            # Map multipliers back to the original time series
            aligned_multipliers = period_ids.map(period_multipliers)

            # Calculate final wealth index
            wealth_index = initial_inv * aligned_multipliers * portfolio_wealth_normalized

        return wealth_index

    def return_ror_ts_ef(self, weights: Union[list, np.ndarray], ror: pd.DataFrame) -> pd.Series:
        """
        Return monthly rate of return time series of rebalanced portfolio given returns time series of the assets.

        Simple version without conditional rebalancing (abs_deviation and rel_deviation are not used).
        This method is used for efficient frontier optimization.
        Default rebalancing period is a Year (end of year).
        For not rebalanced portfolio set period to 'none'.

        Parameters
        ----------
        weights : list of float or np.ndarray
            The target weights for assets in the portfolio.

        ror : pd.DataFrame
            Assets rate of return monthly time series.

        Returns
        -------
        pd.Series
            The monthly rate of return time series of rebalanced portfolio.
        """
        # define data of the first period
        first_date = ror.index[0]
        return_first_period = ror.iloc[0] @ weights

        wealth_index = self.wealth_ts_ef(weights, ror)
        portfolio_ror = wealth_index.pct_change()
        portfolio_ror.loc[first_date] = return_first_period  # replaces NaN with the first period return
        return portfolio_ror



from __future__ import annotations

import math
from typing import Optional, Union

import pandas as pd

import okama.portfolios.core as core
from okama import settings
from okama.common import validators


class CashFlow:
    """
    Parent class for cash flow strategies.

    Parameters
    ----------
    parent : Portfolio
        Parent Portfolio instance.
    """

    NAME = "cash_flow"

    def __init__(
        self,
        parent: core.Portfolio,
        frequency: Optional[str] = "none",
        initial_investment: float = 1000.0,
        time_series_dic: dict = {},
        time_series_discounted_values: bool = False,
    ):
        """
        Initialize the CashFlow strategy.

        Parameters
        ----------
        parent : Portfolio
            The parent Portfolio instance.
        frequency : str, optional
            Frequency of cash flows. Default is "none".
        initial_investment : float, optional
            Initial investment amount. Default is 1000.0.
        time_series_dic : dict, optional
            Dictionary with dates and cash flow values. Default is empty dict.
        time_series_discounted_values : bool, optional
            If True, values in time_series_dic are considered as discounted (PV). Default is False.
        """
        self.parent = parent
        self._frequency = frequency
        self._initial_investment = initial_investment
        self._pandas_frequency = settings.frequency_mapping.get(self.frequency)
        self.time_series = pd.Series(dtype=float)
        self.time_series_dic = time_series_dic
        self.time_series_discounted_values = time_series_discounted_values

    @property
    def frequency(self) -> str:
        """
        The frequency of regular withdrawals or contributions in the strategy.

        Allowed values for frequency:

        - 'none' no frequency (default value)
        - 'year' annual cash flows
        - 'half-year' 6 months cash flows
        - 'quarter' 3 months cash flows
        - 'month' 1 month cash flows

        Returns
        -------
        str
            The frequency of withdrawals or contributions.
        """
        return self._frequency

    @frequency.setter
    def frequency(self, frequency):
        if frequency in settings.frequency_mapping.keys():
            self._clear_cf_cache()
            self._frequency = frequency
            self._pandas_frequency = settings.frequency_mapping.get(self.frequency)
        else:
            raise ValueError(f"frequency must be in {settings.frequency_mapping.keys()}")

    @property
    def periods_per_year(self) -> int:
        """
        Get the number of periods per year based on the frequency.

        Returns
        -------
        int
            Number of periods per year.
        """
        return settings.frequency_periods_per_year[self.frequency]

    @property
    def initial_investment(self) -> float:
        """
        Initial investment (FV) amount at the start of the calculation period.

        For historical backtesting, this is the investment at `first_date`.
        For Monte Carlo simulations, this is the investment at `last_date`.
        Initial investment must be positive.

        Returns
        -------
        float
            Portfolio initial investment.
        """
        return self._initial_investment

    @initial_investment.setter
    def initial_investment(self, initial_investment):
        if initial_investment is not None:
            validators.validate_real("initial_investment", initial_investment)
            if initial_investment <= 0:
                raise ValueError("Initial investment must be positive.")
        self._clear_cf_cache()
        self._initial_investment = initial_investment

    @property
    def time_series_dic(self) -> dict:
        """
        Cash flow time series in form of dictionary.

        Negative number corresponds to withdrawals, positive number corresponds to contributions.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(first_date="2015-01", last_date="2024-10")  # create Portfolio with default parameters
        >>> # create simple dictionary with cash flow amounts and dates
        >>> d = {"2018-02": 2_000, "2024-03": -4_000}
        >>> ts = ok.TimeSeriesStrategy(pf)  # create TimeSeresStrategy linked to the portfolio
        >>> ts.time_series_dic = d  # use the dictionary to set cash flow
        >>> ts.initial_investment = 1_000  # add initial investments size (optional)
        >>> # Assign the strategy to Portfolio
        >>> pf.dcf.cashflow_parameters = ts
        >>> # Plot wealth index with cash flow
        >>> pf.dcf.wealth_index(discounting="fv", include_negative_values=False).plot()
        >>> plt.show()
        """
        # TODO: rename to extra_cashflow_ts
        return self._time_series_dic

    @time_series_dic.setter
    def time_series_dic(self, time_series_dic):
        self._clear_cf_cache()
        if isinstance(time_series_dic, dict):
            self._time_series_dic = time_series_dic
        else:
            raise TypeError("time_series_dic must be a dictionary.")
        self._make_series_from_dic()

    def _make_series_from_dic(self):
        """
        Create cash flow time series in form of Pandas.Series.
        """
        self.time_series = pd.Series(self._time_series_dic)
        self.time_series.index = pd.to_datetime(self.time_series.index).to_period("M")
        self.time_series.sort_index(inplace=True)
        self.time_series.name = "cashflow_ts"

    def _clear_cf_cache(self):
        if hasattr(self.parent, "dcf"):
            self.parent.dcf._monte_carlo_wealth_fv = pd.DataFrame(dtype=float)
            self.parent.dcf._wealth_index_fv = pd.DataFrame(dtype=float)
            self.parent.dcf._cash_flow_fv = pd.DataFrame(dtype=float)
            self.parent.dcf._monte_carlo_cash_flow_fv = pd.DataFrame(dtype=float)


class IndexationStrategy(CashFlow):
    """
    Cash flow strategy with regular indexed withdrawals or contributions.

    Parameters
    ----------
    parent : Portfolio
        Parent Portfolio instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> pf = ok.Portfolio(first_date="2015-01", last_date="2024-10")  # create Portfolio with default parameters
    >>> # Set the cash flow strategy
    >>> ind = ok.IndexationStrategy(pf) # create IndexationStrategy linked to the portfolio
    >>> ind.initial_investment = 10_000  # add initial investments size
    >>> ind.frequency = "year"  # set cash flow frequency
    >>> ind.amount = -1_500  # set withdrawal size
    >>> ind.indexation = "inflation"
    >>> # Assign the strategy to Portfolio
    >>> pf.dcf.cashflow_parameters = ind
    >>> # Plot wealth index with cash flow
    >>> pf.dcf.wealth_index(discounting="fv", include_negative_values=False).plot()
    >>> plt.show()
    """

    NAME = "fixed_amount"

    def __init__(
        self,
        parent: core.Portfolio,
        frequency: Optional[str] = "none",
        initial_investment: float = 1000.0,
        time_series_dic: dict = {},
        time_series_discounted_values: bool = False,
        amount: float = 0,
        indexation: Optional[Union[str, float]] = None,
    ):
        """
        Initialize the IndexationStrategy.

        Parameters
        ----------
        parent : Portfolio
            The parent Portfolio instance.
        frequency : str, optional
            Frequency of cash flows. Default is "none".
        initial_investment : float, optional
            Initial investment amount. Default is 1000.0.
        time_series_dic : dict, optional
            Dictionary with dates and cash flow values. Default is empty dict.
        time_series_discounted_values : bool, optional
            If True, values in time_series_dic are considered as discounted (PV). Default is False.
        amount : float, optional
            Amount of regular cash flow. Negative for withdrawals, positive for contributions. Default is 0.
            The frequency of withdrawals or contributions is determined by the `frequency` parameter.
        indexation : str or float, optional
            Indexation rate (CAGR). Can be 'inflation' (if portfolio has inflation data) or a float value. Default is None.
        """
        super().__init__(
            parent,
            frequency=frequency,
            initial_investment=initial_investment,
            time_series_dic=time_series_dic,
            time_series_discounted_values=time_series_discounted_values,
        )
        self.portfolio = self.parent
        self._amount = amount
        self.indexation = indexation

    def __repr__(self):
        dic = {
            "Strategy name": self.NAME,
            "Portfolio symbol": self.parent.symbol,
            "Cash flow initial investment": self.initial_investment,
            "Cash flow frequency": self.frequency,
            "Cash flow strategy": self.NAME,
            "Cash flow amount": self.amount,
            "Cash flow indexation": self.indexation,
        }
        return repr(pd.Series(dic))

    @property
    def amount(self):
        """
        Portfolio regular withdrawal or contribution size.
        Negative value corresponds to withdrawals. Positive value corresponds to contributions.
        Cash flow value is indexed each period by 'indexation'.

        The frequency of withdrawals or contributions is determined by the `frequency` parameter.

        Returns
        -------
        float
            Portfolio regular cash flow size.
        """
        return self._amount

    @amount.setter
    def amount(self, amount):
        self._clear_cf_cache()
        validators.validate_real("amount", amount)
        if amount < 0 and abs(amount) > self.initial_investment:
            raise ValueError("It's not possible to withdraw more than the initial investment.")
        self._amount = amount

    @property
    def indexation(self) -> float:
        """
        Portfolio cash flow indexation rate.

        Returns
        -------
        float
            Cash flow indexation rate.
        """
        return self._indexation

    @indexation.setter
    def indexation(self, indexation: Optional[float]):
        if indexation in [None, "inflation"] and hasattr(self.portfolio, "inflation"):
            self._indexation = self.portfolio.get_cagr().loc[self.portfolio.inflation]
        elif indexation == "inflation" and not hasattr(self.portfolio, "inflation"):
            raise ValueError("There is no information about historical inflation. Set inflation=True to calculate.")
        elif indexation is None and not hasattr(self.portfolio, "inflation"):
            self._indexation = settings.DEFAULT_DISCOUNT_RATE
        else:
            validators.validate_real("indexation", indexation)
            self._clear_cf_cache()
            self._indexation = indexation


class PercentageStrategy(CashFlow):
    """
    Cash flow strategy with regular fixed percentage withdrawals or contributions.

    Parameters
    ----------
    parent : Portfolio
        The parent Portfolio instance.
    frequency : str, optional
        Frequency of cash flows. Default is "none".
    initial_investment : float, optional
        Initial investment amount. Default is 1000.0.
    time_series_dic : dict, optional
        Dictionary with dates and cash flow values. Default is empty dict.
    time_series_discounted_values : bool, optional
        If True, values in time_series_dic are considered as discounted (PV). Default is False.
    percentage : float, optional
        Percentage of portfolio balance to be withdrawn or contributed. Negative for withdrawals. Default is 0.0.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> pf = ok.Portfolio(first_date="2015-01", last_date="2024-10")  # create Portfolio with default parameters
    >>> pc = ok.PercentageStrategy(pf)  # create PercentageStrategy linked to the portfolio
    >>> pc.initial_investment = 10_000  # add initial investments size
    >>> pc.frequency = "year"  # set cash flow frequency
    >>> pc.percentage = -0.12  # set withdrawal percentage
    >>> # Assign the strategy to Portfolio
    >>> pf.dcf.cashflow_parameters = pc
    >>> # Plot wealth index with cash flow
    >>> pf.dcf.wealth_index(discounting="fv", include_negative_values=False).plot()
    >>> plt.show()
    """

    NAME = "fixed_percentage"

    def __init__(
        self,
        parent: core.Portfolio,
        frequency: Optional[str] = "none",
        initial_investment: float = 1000.0,
        time_series_dic: dict = {},
        time_series_discounted_values: bool = False,
        percentage: float = 0.0,
    ):
        super().__init__(
            parent,
            frequency=frequency,
            initial_investment=initial_investment,
            time_series_dic=time_series_dic,
            time_series_discounted_values=time_series_discounted_values,
        )
        self.portfolio = self.parent
        self._percentage = percentage

    def __repr__(self):
        dic = {
            "Strategy name": self.NAME,
            "Portfolio symbol": self.parent.symbol,
            "Cash flow initial investment": self.initial_investment,
            "Cash flow frequency": self.frequency,
            "Cash flow strategy": self.NAME,
            "Cash flow percentage": self.percentage,
        }
        return repr(pd.Series(dic))

    @property
    def percentage(self) -> float:
        """
        The percentage of withdrawals or contributions.

        The size of withdrawals or contribution is defined as a percentage of portfolio balance per year.

        Returns
        -------
        float
            The percentage of withdrawals or contributions.
        """
        return self._percentage

    @percentage.setter
    def percentage(self, percentage):
        self._clear_cf_cache()
        validators.validate_real("percentage", percentage)
        if percentage < -1:
            raise ValueError("Withdrawal Percentage must less or equal to the Initial investment (100%).")
        self._percentage = percentage


class TimeSeriesStrategy(CashFlow):
    """
    Cash flow strategy with user-defined withdrawals and contributions.

    Withdrawals, contributions, as well as their dates, are defined in the dictionary.

    Parameters
    ----------
    parent : Portfolio
        The parent Portfolio instance.
    initial_investment : float, optional
        Initial investment amount. Default is 0.
    time_series_dic : dict, optional
        Dictionary with dates and cash flow values. Default is empty dict.
    time_series_discounted_values : bool, optional
        If True, values in time_series_dic are considered as discounted (PV). Default is False.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> pf = ok.Portfolio(first_date="2015-01", last_date="2024-10")  # create Portfolio with default parameters
    >>> # create simple dictionary with cash flow amounts and dates
    >>> d = {"2018-02": 2_000, "2024-03": -4_000}
    >>> ts = ok.TimeSeriesStrategy(pf)  # create TimeSeresStrategy linked to the portfolio
    >>> ts.time_series_dic = d  # use the dictionary to set cash flow
    >>> ts.initial_investment = 1_000  # add initial investments size (optional)
    >>> # Assign the strategy to Portfolio
    >>> pf.dcf.cashflow_parameters = ts
    >>> # Plot wealth index with cash flow
    >>> pf.dcf.wealth_index(discounting="fv", include_negative_values=False).plot()
    >>> plt.show()
    """

    NAME = "time_series"

    def __init__(
        self,
        parent: core.Portfolio,
        initial_investment: float = 0,
        time_series_dic: dict = {},
        time_series_discounted_values: bool = False,
    ):
        super().__init__(
            parent,
            frequency="none",
            initial_investment=initial_investment,
            time_series_dic=time_series_dic,
            time_series_discounted_values=time_series_discounted_values,
        )
        self.portfolio = self.parent

    def __repr__(self):
        dic = {
            "Strategy name": self.NAME,
            "Portfolio symbol": self.portfolio.symbol,
            "Cash flow initial investment": self.initial_investment,
            "Cash flow strategy": self.NAME,
        }
        return repr(pd.Series(dic))


class VanguardDynamicSpending(PercentageStrategy):
    """
    Vanguard Dynamic Spending strategy.

    The withdrawal amount is calculated as a percentage of the portfolio balance.

    The `floor_ceiling` parameter limits the withdrawals relative to the
    previous year's cash flow.

    The absolute withdrawal amount can be optionally limited with
    `min_max_annual_withdrawal`. If `adjust_min_max=True`, these bounds are indexed
    using `indexation`.

    Parameters
    ----------
    parent : Portfolio
        The parent Portfolio instance.
    initial_investment : float, optional
        Initial investment amount. Default is 1000.0.
    time_series_dic : dict, optional
        Dictionary with dates and cash flow values. Default is empty dict.
    time_series_discounted_values : bool, optional
        If True, values in time_series_dic are considered as discounted (PV). Default is False.
    percentage : float, optional
        Percentage of portfolio balance to be withdrawn. Negative value. Default is 0.0.
    min_max_annual_withdrawals : tuple[float, float], optional
        Optional absolute min/max annual withdrawal amounts (positive values). Default is None.
    adjust_min_max : bool, optional
        If True, min/max bounds are indexed using `indexation`. Default is True.
    floor_ceiling : tuple[float, float], optional
        Year-to-year withdrawal change limits relative to the previous year's withdrawal.
        Example: (-0.025, 0.05) means the next withdrawal cannot be more than 2.5% lower
        or 5% higher than the previous year's withdrawal. Default is None.
    adjust_floor_ceiling : bool, optional
        If True, the previous year's withdrawal amount is indexed before applying
        floor/ceiling limits. Default is False.
    indexation : str or float, optional
        Indexation rate. Default is None.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> pf = ok.Portfolio(
    ...     ["SPY.US", "BND.US"],
    ...     weights=[.6, .4],
    ...     first_date="2010-01",
    ...     last_date="2024-10",
    ...     ccy="USD",
    ...     inflation=True
    ... )
    >>> vds = ok.VanguardDynamicSpending(
    ...     parent=pf,
    ...     initial_investment=1_000_000,
    ...     percentage=-0.08,
    ...     floor_ceiling=(-.025, .05),
    ...     min_max_annual_withdrawals=(40_000, 100_000),
    ...     adjust_min_max=True,
    ...     indexation="inflation",
    ... )
    >>> pf.dcf.cashflow_parameters = vds
    >>> pf.dcf.wealth_index(discounting="fv", include_negative_values=False).plot()
    >>> plt.show()
    """

    NAME = "VDS"

    def __init__(
        self,
        parent: core.Portfolio,
        initial_investment: float = 1000.0,
        time_series_dic: dict = {},
        time_series_discounted_values: bool = False,
        percentage: float = 0.0,
        min_max_annual_withdrawals: Optional[tuple[float, float]] = None,
        adjust_min_max: bool = True,
        floor_ceiling: Optional[tuple[float, float]] = None,
        adjust_floor_ceiling: bool = False,
        indexation: Optional[Union[str, float]] = None,
    ):
        """
        Initialize the VanguardDynamicSpending strategy.


        """
        super().__init__(
            parent=parent,
            frequency="year",
            initial_investment=initial_investment,
            time_series_dic=time_series_dic,
            time_series_discounted_values=time_series_discounted_values,
            percentage=percentage,
        )
        self.portfolio = self.parent
        self._min_max_annual_withdrawals = min_max_annual_withdrawals
        self._adjust_min_max = adjust_min_max
        self._floor_ceiling = floor_ceiling
        self.adjust_floor_ceiling = adjust_floor_ceiling
        self.indexation = indexation

    def __repr__(self):
        dic = {
            "Strategy name": self.NAME,
            "Portfolio symbol": self.parent.symbol,
            "Cash flow initial investment": self.initial_investment,
            "Cash flow frequency": self.frequency,
            "Cash flow strategy": self.NAME,
            "Cash flow percentage": self.percentage,  # negative
            "Minimum annual withdrawal": (
                self.min_max_annual_withdrawals[0] if self.min_max_annual_withdrawals is not None else None
            ),  # positive
            "Maximum annual withdrawal": (
                self.min_max_annual_withdrawals[1] if self.min_max_annual_withdrawals is not None else None
            ),  # positive
            "Max and Min withdrawals are indexed": str(self.adjust_min_max),
            "Floor": self.floor_ceiling[0] if self.floor_ceiling is not None else None,  # negative
            "Ceiling": self.floor_ceiling[1] if self.floor_ceiling is not None else None,  # positive
            "Floor and Ceiling are indexed": str(self.adjust_floor_ceiling),
            "Indexation": self.indexation,
        }
        return repr(pd.Series(dic))

    @property
    def percentage(self) -> float:
        """
        The percentage of withdrawals (no contributions are allowed in VDS strategy). The value must be negative.

        The size of withdrawals is defined as a percentage of portfolio balance per year.

        Returns
        -------
        float
            The percentage of withdrawals or contributions.
        """
        return self._percentage

    @percentage.setter
    def percentage(self, percentage):
        self._clear_cf_cache()
        validators.validate_real("percentage", percentage)
        if percentage < -1:
            raise ValueError("Withdrawal Percentage must less or equal to the Initial investment (100%).")
        if percentage > 0:
            raise ValueError("Only withdrawals are allowed in VDS strategy. Percentage must be negative or zero.")
        self._percentage = percentage

    @property
    def frequency(self):
        """
        Frequency of cash flows. Always 'year' for VDS.
        """
        return "year"

    @frequency.setter
    def frequency(self, value):
        if value != "year":
            raise AttributeError("In VDS the 'frequency' can only be equal to a year.")
        else:
            CashFlow.frequency.fset(self, "year")

    @property
    def min_max_annual_withdrawals(self):
        """
        Minimum and maximum annual withdrawal limits.
        """
        return self._min_max_annual_withdrawals

    @min_max_annual_withdrawals.setter
    def min_max_annual_withdrawals(self, value: Optional[tuple[float, float]]):
        if not isinstance(value, tuple):
            raise TypeError("min_max_annual_withdrawals must be a tuple (float, float).")
        min_w = value[0]
        max_w = value[1]
        validators.validate_real("minimum annual withdrawal", min_w)
        validators.validate_real("maximum annual withdrawal", max_w)
        if min_w < 0:
            raise ValueError("Min withdrawal cannot be negative.")
        if max_w < 0:
            raise ValueError("Max withdrawal cannot be negative.")
        if min_w > max_w:
            raise ValueError("Minimum withdrawal cannot be greater than maximum withdrawal.")
        self._clear_cf_cache()
        self._min_max_annual_withdrawals = value

    @property
    def adjust_min_max(self):
        """
        Whether to adjust min and max limits by indexation.
        """
        return self._adjust_min_max

    @adjust_min_max.setter
    def adjust_min_max(self, value):
        if not isinstance(value, bool):
            raise TypeError("adjust_min_max must be a True or False.")
        self._clear_cf_cache()
        self._adjust_min_max = value

    @property
    def floor_ceiling(self):
        """
        Floor and ceiling percentages relative to the previous year's withdrawal.
        """
        return self._floor_ceiling

    @floor_ceiling.setter
    def floor_ceiling(self, value: Optional[tuple[float, float]]):
        if not isinstance(value, tuple):
            raise TypeError("floor_ceiling must be a tuple (float, float).")
        floor = value[0]
        ceiling = value[1]
        validators.validate_real("floor", floor)
        validators.validate_real("ceiling", ceiling)
        if floor >= 0:
            raise ValueError("Floor must be negative.")
        if ceiling <= 0:
            raise ValueError("Ceiling must be positive.")
        self._clear_cf_cache()
        self._floor_ceiling = value

    @property
    def adjust_floor_ceiling(self):
        """
        Whether to adjust floor and ceiling by indexation.
        """
        return self._adjust_floor_ceiling

    @adjust_floor_ceiling.setter
    def adjust_floor_ceiling(self, value):
        if not isinstance(value, bool):
            raise TypeError("adjust_min_max must be a True or False.")
        self._clear_cf_cache()
        self._adjust_floor_ceiling = value

    @property
    def indexation(self) -> float:
        """
        Indexation rate for Minimum/Maximum annual withdrawal and Floor/Ceiling.

        Returns
        -------
        float
            Indexation rate.
        """
        return self._indexation

    @indexation.setter
    def indexation(self, indexation: Optional[float]):
        if indexation in [None, "inflation"] and hasattr(self.portfolio, "inflation"):
            self._indexation = self.portfolio.get_cagr().loc[self.portfolio.inflation]
        elif indexation == "inflation" and not hasattr(self.portfolio, "inflation"):
            raise ValueError("There is no information about historical inflation. Set inflation=True to calculate.")
        elif indexation is None and not hasattr(self.portfolio, "inflation"):
            self._indexation = settings.DEFAULT_DISCOUNT_RATE
        else:
            validators.validate_real("indexation", indexation)
            self._clear_cf_cache()
            self._indexation = indexation

    def _calculate_withdrawal_size(self, last_withdrawal: float, balance: float, number_of_periods: int) -> float:
        """
        Calculate regular withdrawal size (Extra Withdrawals are not taken into account). Used in helpers.

        Parameters
        ----------
        last_withdrawal : float
            Withdrawal amount from the previous period.
        balance : float
            Current portfolio balance.
        number_of_periods : int
            Number of periods elapsed.

        Returns
        -------
        float
            Calculated withdrawal size.
        """
        # All values are positive
        withdrawal_size_by_percentage = balance * abs(self.percentage)
        if self.floor_ceiling is not None:
            floor, ceiling = self.floor_ceiling
            floor_indexed = (
                abs(last_withdrawal) * (1 + self.indexation) * (1 + floor)
                if self.adjust_floor_ceiling
                else abs(last_withdrawal) * (1 + floor)
            )
            ceiling_indexed = (
                abs(last_withdrawal) * (1 + self.indexation) * (1 + ceiling)
                if self.adjust_floor_ceiling
                else abs(last_withdrawal) * (1 + ceiling)
            )
        if self.min_max_annual_withdrawals is not None:
            min_withdrawal, max_withdrawal = self.min_max_annual_withdrawals
            min_indexed = (
                abs(min_withdrawal) * (1 + self.indexation) ** number_of_periods
                if self.adjust_min_max
                else abs(min_withdrawal)
            )
            max_indexed = (
                abs(max_withdrawal) * (1 + self.indexation) ** number_of_periods
                if self.adjust_min_max
                else abs(max_withdrawal)
            )
        # Chek what limitation is actual
        if self.floor_ceiling is not None and self.min_max_annual_withdrawals is not None:
            # Upper limit
            if ceiling_indexed > max_indexed:
                max_final = max_indexed
            elif min_indexed < ceiling_indexed <= max_indexed:
                max_final = ceiling_indexed
            else:
                # floor_indexed & ceiling_indexed = 0 for the first withdrawal (last_withdrawal = 0)
                max_final = max_indexed
            # Lower limit
            if floor_indexed > min_indexed:
                min_final = floor_indexed
            elif 0 < floor_indexed <= min_indexed:
                min_final = min_indexed
            else:
                # floor_indexed & ceiling_indexed = 0 for the first withdrawal (last_withdrawal = 0)
                min_final = min_indexed
        elif self.floor_ceiling is None and self.min_max_annual_withdrawals is not None:
            # print("ceiling is None")
            min_final = min_indexed
            max_final = max_indexed
        elif self.floor_ceiling is not None and self.min_max_annual_withdrawals is None:
            # print("min_max_annual_withdrawals is None")
            min_final = floor_indexed
            max_final = ceiling_indexed if ceiling_indexed != 0 else withdrawal_size_by_percentage
        else:
            # no limits
            min_final = -math.inf
            max_final = math.inf
        # Apply the limitation to the withdrawal
        if min_final <= withdrawal_size_by_percentage <= max_final:
            withdrawal = -withdrawal_size_by_percentage
            # print(f"withdrawal by percentage. Max: {max_final: .0f}, Min: {min_final: .0f}")
        elif withdrawal_size_by_percentage > max_final:
            withdrawal = -max_final
            # print(f"withdrawal by max_final. By percentage was {withdrawal_size_by_percentage: .0f}")
        elif withdrawal_size_by_percentage < min_final:
            withdrawal = -min_final
            # print(f"withdrawal by min_final. By percentage was {withdrawal_size_by_percentage: .0f}")
        else:
            raise ValueError("Wrong withdrawal size. Check the calculation.")
        return withdrawal


class CutWithdrawalsIfDrawdown(IndexationStrategy):
    """
    Withdrawal strategy that reduces the withdrawal amount if the portfolio drawdown exceeds a certain threshold.

    The reduction coefficients are defined in `crash_threshold_reduction` list of tuples.

    Parameters
    ----------
    parent : Portfolio
        The parent Portfolio instance.
    frequency : str, optional
        Frequency of cash flows. Default is "year".
    initial_investment : float, optional
        Initial investment amount. Default is 1000.0.
    time_series_dic : dict, optional
        Dictionary with dates and cash flow values. Default is empty dict.
    time_series_discounted_values : bool, optional
        If True, values in time_series_dic are considered as discounted (PV). Default is False.
    amount : float, optional
        Regular withdrawal amount (negative value) before reduction. Default is 0.0.
        The frequency of withdrawals is determined by the `frequency` parameter.
    indexation : str or float, optional
        Indexation rate for the withdrawal amount. Default is None.
    crash_threshold_reduction : list[tuple[float, float]], optional
        List of tuples (threshold, reduction_coefficient).
        Example: [(0.20, 0.40)] means if drawdown > 20%, reduce withdrawal by 40%.
        Default is [(.20, .40), (.50, 1)].

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> pf = ok.Portfolio(
    ...     ["SPY.US", "BND.US"],
    ...     weights=[.6, .4],
    ...     first_date="2010-01",
    ...     last_date="2024-10",
    ...     ccy="USD",
    ...     inflation=True,
    ... )
    >>> cwid = ok.CutWithdrawalsIfDrawdown(
    ...     parent=pf,
    ...     initial_investment=1_000_000,
    ...     frequency="year",
    ...     amount=-60_000,
    ...     indexation="inflation",
    ...     crash_threshold_reduction=[(.10, .25), (.20, .50), (.35, 1)],
    ... )
    >>> pf.dcf.cashflow_parameters = cwid
    >>> pf.dcf.wealth_index(discounting="fv", include_negative_values=False).plot()
    >>> plt.show()
    """

    NAME = "CWD"

    def __init__(
        self,
        parent: core.Portfolio,
        frequency: Optional[str] = "year",
        initial_investment: float = 1000.0,
        time_series_dic: dict = {},
        time_series_discounted_values: bool = False,
        amount: float = 0.0,
        indexation: Optional[Union[str, float]] = None,
        crash_threshold_reduction: list[tuple[float, float]] = [
            (0.20, 0.40),
            (0.50, 1),
        ],
    ):
        super().__init__(
            parent=parent,
            frequency=frequency,
            initial_investment=initial_investment,
            time_series_dic=time_series_dic,
            time_series_discounted_values=time_series_discounted_values,
            amount=amount,
            indexation=indexation,
        )
        self._crash_threshold_reduction_series = None
        self.portfolio = self.parent
        self._crash_threshold_reduction = crash_threshold_reduction
        self._crash_threshold_reduction_series = self.make_series_from_list(self.crash_threshold_reduction)

    def __repr__(self):
        dic = {
            "Strategy name": self.NAME,
            "Portfolio symbol": self.parent.symbol,
            "Cash flow initial investment": self.initial_investment,
            "Cash flow frequency": self.frequency,
            "Cash flow strategy": self.NAME,
            "Cash flow amount": self.amount,
            "Cash flow indexation": self.indexation,
            "Crash threshold reduction": self.crash_threshold_reduction,
        }
        return repr(pd.Series(dic))

    @property
    def amount(self):
        """
        Portfolio regular withdrawals size. Must be negative.
        Cash flow value is indexed each period by 'indexation'.

        The frequency of withdrawals or contributions is determined by the `frequency` parameter.

        Returns
        -------
        float
            Portfolio regular cash flow size.
        """
        return self._amount

    @amount.setter
    def amount(self, amount):
        self._clear_cf_cache()
        validators.validate_real("amount", amount)
        if abs(amount) > self.initial_investment:
            raise ValueError("It's not possible to withdraw more than the initial investment.")
        if amount > 0:
            raise ValueError("Amount must be negative or zero for withdrawals in CWD strategy.")
        self._amount = amount

    @property
    def crash_threshold_reduction(self):
        """
        List of tuples (threshold, reduction_coefficient) that define how much withdrawals are reduced
        based on the portfolio drawdown depth.

        Example
        -------
        >>> pf = ok.Portfolio(["SPY.US", "BND.US"], weights=[0.6, 0.4], ccy="USD")
        >>> strategy = ok.CutWithdrawalsIfDrawdown(parent=pf)
        >>> strategy.crash_threshold_reduction = [(0.05, 0.20), (0.10, 0.40), (0.20, 0.50), (0.30, 1.0)]
        >>> strategy.crash_threshold_reduction
        [(0.05, 0.2), (0.1, 0.4), (0.2, 0.5), (0.3, 1.0)]

        If drawdown is 5%, reduce withdrawal by 20%; if 10%, reduce by 40%; if 20%, reduce by 50%.
        If drawdown is 30% or more, withdrawals stop entirely (100% reduction).
        """
        return self._crash_threshold_reduction

    @crash_threshold_reduction.setter
    def crash_threshold_reduction(self, value):
        self._clear_cf_cache()
        self._crash_threshold_reduction_series = self.make_series_from_list(value)
        for threshold, reduction in self._crash_threshold_reduction_series.items():
            validators.validate_real("threshold", threshold)
            validators.validate_real("reduction", reduction)
            if abs(threshold) >= 1 or threshold == 0:
                raise ValueError("crash_threshold_reduction first values (threshold) must be in the interval (0, 1).")
            if abs(reduction) > 1:
                raise ValueError("crash_threshold_reduction second values (reductiuon) must be in the interval [0, 1].")
        self._crash_threshold_reduction = value

    def _calculate_withdrawal_size(self, drawdown: float, withdrawal_without_drawdowns: float) -> float:
        """
        Calculate regular withdrawal size (Extra Withdrawals are not taken into account). Used in helpers.

        Parameters
        ----------
        drawdown : float
            Current portfolio drawdown (negative value).
        withdrawal_without_drawdowns : float
            Withdrawal amount calculated without considering drawdowns.

        Returns
        -------
        float
            Calculated withdrawal size.
        """
        withdrawal = abs(withdrawal_without_drawdowns)
        for threshold, reduction in self._crash_threshold_reduction_series.items():
            if abs(drawdown) >= threshold:
                withdrawal *= 1 - reduction
                break
        return -withdrawal

    def make_series_from_list(self, l: list[tuple[float, float]]) -> pd.Series:
        """
        Convert list of tuples to pandas Series.

        Parameters
        ----------
        l : list[tuple[float, float]]
            List of tuples.

        Returns
        -------
        pd.Series
            Series with threshold as index and reduction as value.
        """
        indices = [abs(index) for index, _ in l]
        values = [abs(value) for _, value in l]
        crash_series = pd.Series(values, index=indices)
        return crash_series.sort_index(ascending=False)



import json
from typing import Optional
from functools import lru_cache

import pandas as pd

from okama.api import api_methods, namespaces


@lru_cache()
def search(search_string: str, namespace: Optional[str] = None, response_format: str = "frame") -> json:
    """
    Search symbols by ticker, name, or ISIN.

    When ``namespace`` is provided, the search is performed within the cached
    table returned by ``ok.symbols_in_namespace(namespace)``. Otherwise the
    query is delegated to the API search endpoint across all namespaces.

    Parameters
    ----------
    search_string : str
        Case-insensitive text used to match symbol names, tickers, and ISINs.

    namespace : str, optional
        Namespace code such as ``"US"`` or ``"XETR"``. If omitted, all
        available namespaces are searched.

    response_format : {'frame', 'json'}, default 'frame'
        Format of the returned search results.

    Returns
    -------
    pandas.DataFrame or str or list
        Search results.

        - Returns a ``DataFrame`` when ``response_format='frame'``.
        - Returns a JSON string in pandas ``records`` orientation when
          ``namespace`` is provided and ``response_format='json'``.
        - Returns the parsed API JSON payload as a list when ``namespace`` is
          omitted and ``response_format='json'``.

    Raises
    ------
    ValueError
        If ``response_format`` is not ``'frame'`` or ``'json'``.

    Examples
    --------
    >>> result = ok.search("SPY", namespace="US")
    >>> result.empty
    False
    >>> {"symbol", "ticker", "name"}.issubset(result.columns)
    True
    """
    # search for string in a single namespace
    if namespace:
        df = namespaces.symbols_in_namespace(namespace.upper())
        condition1 = df["name"].str.contains(search_string, case=False)
        condition2 = df["ticker"].str.contains(search_string, case=False)
        condition3 = df["isin"].str.contains(search_string, case=False)
        frame_response = df[condition1 | condition2 | condition3]
        if response_format.lower() == "frame":
            return frame_response
        elif response_format.lower() == "json":
            return frame_response.to_json(orient="records")
        else:
            raise ValueError('response_format must be "json" or "frame"')
    # search for string in all namespaces
    string_response = api_methods.API.search(search_string)
    json_response = json.loads(string_response)
    if response_format.lower() == "frame":
        df = pd.DataFrame(json_response[1:], columns=json_response[0])
        return df
    elif response_format.lower() == "json":
        return json_response
    else:
        raise ValueError('response_format must be "json" or "frame"')



from dataclasses import dataclass

import pandas as pd


@dataclass(frozen=True)
class Result:
    """
    The result of finding a solution for `find_the_largest_withdrawals_size()`.

    Attributes
    ----------
    success : bool
        Whether or not the solver exited successfully.

    withdrawal_abs : float
        The absolute amount of withdrawal size (the best solution found).
        It is the first withdrawal value (before it was indexed in IndexationStrategy).

    withdrawal_rel : float
        The relative amount of withdrawal size (the best solution found).
        The first withdrawal value (before it was indexed in IndexationStrategy) divided by the initial investment.

    error_rel : float
        Characterizes how accurately the goal is fulfilled. The goal is set in the parameters.

    solutions : pd.DataFrame
        The history of attempts to find solutions (withdrawal values and error level).
    """

    success: bool
    withdrawal_abs: float
    withdrawal_rel: float
    error_rel: float
    solutions: pd.DataFrame(columns=["withdrawal_abs", "withdrawal_rel", "error_rel", "error_rel_change"])

    def __repr__(self):
        dic = {
            "success": self.success,
            "withdrawal_abs": self.withdrawal_abs,
            "withdrawal_rel": self.withdrawal_rel,
            "error_rel": self.error_rel,
            "attempts": self.solutions.shape[0],
        }
        return repr(pd.Series(dic))



from collections import namedtuple

# tickers
default_ticker = "SPY.US"
default_namespace = "US"
default_tickers_list = ["SPY.US", "BND.US"]  # required in single_period.py

# Default Macro Symbols
default_macro_inflation = "RUB.INFL"
default_macro_rate = "RUS_CBR.RATE"
default_macro_indicator = "USA_CAPE10.RATIO"

DEFAULT_DISCOUNT_RATE = 0.05
DEFAULT_INITIAL_INVESTMENT = 1000.0

_MONTHS_PER_YEAR = 12
PeriodLength = namedtuple("PeriodLength", "years months")

# Distributions
distributions = ("norm", "lognorm", "t")

# From Pandas resamples alias: https://pandas.pydata.org/docs/user_guide/timeseries.html#timeseries-offset-aliases
frequency_mapping = {"none": "none", "year": "Y", "half-year": "2Q", "quarter": "Q", "month": "M"}
grouper_frequency_mapping = {"none": "none", "year": "YE", "half-year": "2QE", "quarter": "QE", "month": "ME"}
frequency_periods_per_year = {"none": 0, "year": 1, "half-year": 2, "quarter": 4, "month": 12}
